# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from unittest.mock import patch

from odfdo.cell import Cell
from odfdo.table import Table


def test_get_formatted_text_rst_empty_col_size_0():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "")
    table.set_value("C1", "v3")
    # Manually clear B1
    row = table.get_row(0, clone=False)
    cell = row.get_cell(1, clone=False)
    cell.clear()
    txt = table.get_formatted_text(
        context={
            "rst_mode": True,
            "no_img_level": 0,
        }
    )
    assert "v1" in txt and "v3" in txt


def test_get_formatted_text_rst_min_size():
    table = Table("Test")
    # We need real_size > free_size.
    table.set_value("A1", "w" * 500)
    table.set_value("B1", "w" * 20)
    txt = table.get_formatted_text(
        context={
            "rst_mode": True,
            "no_img_level": 0,
        }
    )
    assert "word" in txt or "w" in txt


def test_get_formatted_text_empty_col():
    table = Table("TestTable")
    table.set_value("B1", "content")
    txt = table.get_formatted_text(
        context={
            "rst_mode": True,
            "no_img_level": 0,
        }
    )
    assert ".." in txt


def test_get_formatted_text_resize():
    table = Table("TestTable")
    long_content = "word" * 100
    table.set_value("A1", long_content)
    table.set_value("B1", "small")
    txt = table.get_formatted_text(
        context={
            "rst_mode": True,
            "no_img_level": 0,
        }
    )
    assert "word" in txt


def test_get_formatted_text_indent_list():
    table = Table("TestTable")
    table.set_value("A1", "- list item")
    table.set_value("B1", ".. another item")
    txt = table.get_formatted_text(
        context={
            "rst_mode": True,
            "no_img_level": 0,
        }
    )
    assert "- list item" in txt
    assert ".. another item" in txt


def test_get_formatted_text_rst_empty_col_size_0_final():
    table = Table("T")
    # A1 has value, B1 is empty, C1 has value
    table.set_value("A1", "v")
    table.set_value("C1", "v")
    # B1 is a real cell but empty
    table.set_cell("B1", Cell())
    txt = table.get_formatted_text(context={"rst_mode": True, "no_img_level": 0})
    assert "v" in txt


def test_get_formatted_text_edge_case():
    original_max = max

    def mock_max(*args, **kwargs):
        if len(args) == 2 and args[0] == 2 and args[1] == 0:
            return 0
        return original_max(*args, **kwargs)

    table = Table("T")
    table.set_value("A1", "v")
    table.set_value("B1", "")  # len 0
    # C1 to keep B1 column
    table.set_value("C1", "v")

    with patch("odfdo.table.max", side_effect=mock_max):
        txt = table.get_formatted_text(
            context={
                "rst_mode": True,
                "no_img_level": 0,
            }
        )
        assert "v" in txt



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo.const import ODF_CONTENT
from odfdo.container import Container
from odfdo.content import Content
from odfdo.style import Style


@pytest.fixture
def example_container(samples) -> Iterable[Container]:
    container = Container()
    container.open(samples("example.odt"))
    yield container


@pytest.fixture
def content(example_container) -> Iterable[Content]:
    yield Content(ODF_CONTENT, example_container)


def test_content_class(content):
    assert isinstance(content, Content)


def test_content_str(content):
    assert str(content) == str(content.body)


def test_get_style_contexts_font_face(content):
    ctxs = content._get_style_contexts("font-face")
    assert len(ctxs) == 1
    assert ctxs[0].tag == "office:font-face-decls"


def test_get_style_contexts_other(content):
    ctxs = content._get_style_contexts("paragraph")
    assert len(ctxs) == 2
    assert ctxs[0].tag == "office:font-face-decls"
    assert ctxs[1].tag == "office:automatic-styles"


def test_get_style_contexts_none(content):
    ctxs = content._get_style_contexts(None)
    assert len(ctxs) == 2
    assert ctxs[0].tag == "office:font-face-decls"
    assert ctxs[1].tag == "office:automatic-styles"


def test_get_styles_all(content):
    styles = content.get_styles()
    assert len(styles) > 0
    tags = {s.tag for s in styles}
    assert "style:font-face" in tags
    assert "style:style" in tags


def test_get_styles_family(content):
    styles = content.get_styles(family="paragraph")
    assert len(styles) > 0
    for s in styles:
        assert s.get_attribute("style:family") == "paragraph"


def test_get_style_by_name(content):
    # P1 should exist in example.odt
    style = content.get_style("paragraph", "P1")
    assert style is not None
    assert style.get_attribute("style:name") == "P1"
    assert style.get_attribute("style:family") == "paragraph"


def test_get_style_font_face(content):
    # Liberation Serif should exist
    style = content.get_style("font-face", "Liberation Serif")
    assert style is not None
    assert style.tag == "style:font-face"
    assert style.get_attribute("style:name") == "Liberation Serif"


def test_get_style_not_found(content):
    style = content.get_style("paragraph", "nonexistent_style_name")
    assert style is None


def test_get_style_by_element(content):
    style = content.get_style("paragraph", "P1")
    assert style is not None
    result = content.get_style("paragraph", style)
    assert result is style


def test_get_style_by_display_name(content):
    # Let's add a style with a display name to test this
    new_style = Style("paragraph", name="MyStyle", display_name="Friendly Name")
    content.get_element("//office:automatic-styles").append(new_style)

    style = content.get_style("paragraph", display_name="Friendly Name")
    assert style is not None
    assert style.get_attribute("style:name") == "MyStyle"
    assert style.get_attribute("style:display-name") == "Friendly Name"


def test_get_styles_missing_context(example_container):
    # Create a content part with missing styles elements
    content = Content(ODF_CONTENT, example_container)
    # Remove font-face-decls and automatic-styles
    ffd = content.get_element("//office:font-face-decls")
    if ffd:
        ffd.delete()
    aus = content.get_element("//office:automatic-styles")
    if aus:
        aus.delete()

    styles = content.get_styles()
    assert len(styles) == 0

    style = content.get_style("paragraph", "P1")
    assert style is None



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from odfdo.table import Row


def test_default():
    row = Row()
    expected = "<table:table-row/>"
    assert row.serialize() == expected


def test_width():
    row = Row(1)
    expected = "<table:table-row><table:table-cell/></table:table-row>"
    assert row.serialize() == expected


def test_repeated():
    row = Row(repeated=3)
    expected = '<table:table-row table:number-rows-repeated="3"/>'
    assert row.serialize() == expected


def test_style():
    row = Row(style="ro1")
    expected = '<table:table-row table:style-name="ro1"/>'
    assert row.serialize() == expected


def test_all():
    row = Row(1, repeated=3, style="ro1")
    expected = (
        '<table:table-row table:number-rows-repeated="3" '
        'table:style-name="ro1">'
        "<table:table-cell/>"
        "</table:table-row>"
    )
    assert row.serialize() == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
from __future__ import annotations

from collections.abc import Iterable
from unittest.mock import patch

import pytest

from odfdo.cell import Cell
from odfdo.document import Document
from odfdo.element import Element
from odfdo.row import Row
from odfdo.table import Table


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield document.body.get_table(name="Example1")


def test_get_cell_alpha(table):
    cell = table.get_cell("D3")
    assert cell.get_value() == 2
    assert cell.text_content == "2"
    assert cell.type == "float"
    assert cell.style == "ce5"
    assert cell.x == 3
    assert cell.y == 2


def test_get_cell_tuple(table):
    cell = table.get_cell((3, 2))
    assert cell.get_value() == 2
    assert cell.text_content == "2"
    assert cell.type == "float"
    assert cell.style == "ce5"
    assert cell.x == 3
    assert cell.y == 2


def test_set_cell_value(table):
    table.set_value("D3", "Changed")
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, "Changed", 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_get_cell_list(table):
    assert len(list(table.get_cells(flat=True))) == 28


def test_get_cell_list_regex(table):
    coordinates = [
        (cell.x, cell.y) for cell in table.get_cells(content=r"3", flat=True)
    ]
    expected = [
        (4, 0),
        (5, 0),
        (6, 0),
        (4, 1),
        (5, 1),
        (6, 1),
        (4, 2),
        (5, 2),
        (6, 2),
        (2, 3),
    ]
    assert coordinates == expected


def test_get_cell_list_style(table):
    coordinates = [
        (cell.x, cell.y) for cell in table.get_cells(style=r"ce5", flat=True)
    ]
    expected = [(1, 1), (5, 1), (3, 2)]
    assert coordinates == expected


def test_insert(table):
    cell = table.insert_cell("B3")
    assert type(cell) is Cell
    assert cell.x == 1
    assert cell.y == 2


def test_insert_cell(table):
    cell = table.insert_cell("B3", Cell("Inserted"))
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, "Inserted", 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7, None],
    ]
    # Test columns are synchronized
    assert table.width == 8
    assert cell.x == 1
    assert cell.y == 2


def test_append(table):
    cell = table.append_cell(1)
    assert type(cell) is Cell
    assert cell.x == table.width - 1


def test_append_cell(table):
    cell = table.append_cell(1, Cell("Appended"))
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 1, 1, 2, 3, 3, 3, "Appended"],
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 2, 3, 4, 5, 6, 7, None],
    ]
    # Test columns are synchronized
    assert table.width == 8
    assert cell.x == table.width - 1


def test_delete_cell(table):
    table.delete_cell("F3")
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, None],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    # Test columns are synchronized
    assert table.width == 7


def test_get_cell_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_row2_base", return_value=None):
        with pytest.raises(ValueError):
            table.get_cell((0, 0))
    _row = table.get_row(0, clone=False)
    with patch.object(Row, "get_cell", return_value=None):
        with pytest.raises(ValueError):
            table.get_cell((0, 0))


def test_get_value_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_row2_base", return_value=None):
        with pytest.raises(ValueError):
            table.get_value((0, 0))


def test_set_cell_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_row2_base", return_value=None):
        with pytest.raises(ValueError):
            table.set_cell((0, 0), Cell("v"))


def test_insert_cell_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_row2_base", return_value=None):
        with pytest.raises(ValueError):
            table.insert_cell((0, 0), Cell("v"))


def test_delete_cell_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_row2_base", return_value=None):
        with pytest.raises(ValueError):
            table.delete_cell((0, 0))


def test_get_values_coord():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    vals = table.get_values(coord="A1:B1")
    assert vals == [["v1", "v2"]]


def test_get_cells_options():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    table.set_value("A2", "v3")
    table.set_value("B2", 123)
    cells = table.get_cells(flat=True)
    assert len(cells) == 4
    cells = table.get_cells(coord="A1:B2", cell_type="string", flat=True)
    assert len(cells) == 3
    cells = table.get_cells(coord="A1:B2")
    assert len(cells) == 2
    assert len(cells[0]) == 2


def test_get_cell_malformed():
    table = Table("T")
    with pytest.raises(ValueError):
        table.get_cell("A")


def test_get_value_malformed():
    table = Table("T")
    with pytest.raises(ValueError):
        table.get_value("1")
    with pytest.raises(ValueError):
        table.get_value("A")


def test_insert_cell_malformed():
    table = Table("T")
    with pytest.raises(ValueError):
        table.insert_cell("1", Cell())
    with pytest.raises(ValueError):
        table.insert_cell("A", Cell())


def test_get_cell_keep_repeated_fix():
    xml = (
        '<table:table table:name="T1" xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0" '
        'xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0" '
        'xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0">'
        "<table:table-row>"
        '<table:table-cell table:number-columns-repeated="5" office:value-type="string">'
        "<text:p>v</text:p>"
        "</table:table-cell>"
        "</table:table-row>"
        "</table:table>"
    )
    table = Element.from_tag(xml)
    c = table.get_cell((0, 0), keep_repeated=True)
    assert c.get_attribute("table:number-columns-repeated") == "5"

    c2 = table.get_cell((0, 0), keep_repeated=False)
    assert c2.get_attribute("table:number-columns-repeated") is None



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Luis Belmar-Letelier <luis@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
from importlib import resources as rso
from io import BytesIO
from unittest.mock import MagicMock, PropertyMock, patch

import pytest

from odfdo.annotation import Annotation
from odfdo.const import (
    ODF_CONTENT,
    ODF_EXTENSIONS,
    ODF_MANIFEST,
    ODF_MANIFEST_RDF,
    ODF_META,
    ODF_SETTINGS,
)
from odfdo.content import Content
from odfdo.document import Document, _get_part_class
from odfdo.element import Element
from odfdo.frame import Frame
from odfdo.header import Header
from odfdo.image import DrawFillImage, DrawImage
from odfdo.manifest import Manifest
from odfdo.master_page import StyleMasterPage
from odfdo.meta import Meta
from odfdo.note import Note
from odfdo.paragraph import Paragraph
from odfdo.settings import Settings
from odfdo.style import Style
from odfdo.style_base import StyleBase
from odfdo.table import Table
from odfdo.utils.blob import Blob
from odfdo.xmlpart import XmlPart


def _copied_template(tmp_path, template):
    src = rso.files("odfdo.templates") / template
    dest = tmp_path / template
    dest.write_bytes(src.read_bytes())
    return dest


def test_non_existing(samples):
    with pytest.raises(IOError):
        Document(samples("notexisting"))


def test_text_template(tmp_path):
    template = _copied_template(tmp_path, "text.ott")
    print(template)
    assert Document.new(template)


def test_spreadsheet_template(tmp_path):
    template = _copied_template(tmp_path, "spreadsheet.ots")
    assert Document.new(template)


def test_presentation_template(tmp_path):
    template = _copied_template(tmp_path, "presentation.otp")
    assert Document.new(template)


def test_drawing_template(tmp_path):
    template = _copied_template(tmp_path, "drawing.otg")
    assert Document.new(template)


def test_mimetype(tmp_path):
    template = _copied_template(tmp_path, "drawing.otg")
    document = Document.new(template)
    mimetype = document.mimetype
    assert "template" not in mimetype
    manifest = document.get_part(ODF_MANIFEST)
    media_type = manifest.get_media_type("/")
    assert "template" not in media_type


def test_bad_type():
    with pytest.raises(IOError):
        Document.new("foobar")


def test_text_type():
    document = Document("text")
    assert document.mimetype == ODF_EXTENSIONS["odt"]


def test_text_type_2():
    document = Document("texte")
    assert document.mimetype == ODF_EXTENSIONS["odt"]


def test_text_type_3():
    document = Document("odt")
    assert document.mimetype == ODF_EXTENSIONS["odt"]


def test_spreadsheet_type():
    document = Document.new("spreadsheet")
    assert document.mimetype == ODF_EXTENSIONS["ods"]


def test_spreadsheet_type_2():
    document = Document.new("tableur")
    assert document.mimetype == ODF_EXTENSIONS["ods"]


def test_spreadsheet_type_3():
    document = Document.new("ods")
    assert document.mimetype == ODF_EXTENSIONS["ods"]


def test_presentation_type():
    document = Document("presentation")
    assert document.mimetype == ODF_EXTENSIONS["odp"]


def test_presentation_type_2():
    document = Document("odp")
    assert document.mimetype == ODF_EXTENSIONS["odp"]


def test_drawing_type():
    document = Document.new("drawing")
    assert document.mimetype == ODF_EXTENSIONS["odg"]


def test_drawing_type_2():
    document = Document.new("odg")
    assert document.mimetype == ODF_EXTENSIONS["odg"]


def test_drawing_type_3():
    document = Document.new("graphics")
    assert document.mimetype == ODF_EXTENSIONS["odg"]


def test_drawing_type_4():
    document = Document.new("graphic")
    assert document.mimetype == ODF_EXTENSIONS["odg"]


def test_case_filesystem(samples):
    assert Document(samples("example.odt"))


def test_case_get_mimetype(samples):
    document = Document(samples("example.odt"))
    assert document.mimetype == ODF_EXTENSIONS["odt"]


def test_case_get_content(samples):
    document = Document(samples("example.odt"))
    content = document.get_part(ODF_CONTENT)
    assert isinstance(content, Content)


def test_case_get_meta(samples):
    document = Document(samples("example.odt"))
    meta = document.get_part(ODF_META)
    assert isinstance(meta, Meta)


def test_case_get_manifest(samples):
    document = Document(samples("example.odt"))
    manifest = document.get_part(ODF_MANIFEST)
    assert isinstance(manifest, Manifest)


def test_case_get_settings(samples):
    document = Document(samples("example.odt"))
    settings = document.get_part(ODF_SETTINGS)
    assert isinstance(settings, Settings)


def test_case_get_body(samples):
    document = Document(samples("example.odt"))
    body = document.body
    assert body.tag == "office:text"


def test_case_clone_body_none(samples):
    document = Document(samples("example.odt"))
    _dummy = document.body
    clone = document.clone
    # new body cache should be empty
    assert clone._Document__body is None


def test_case_clone_xmlparts_empty(samples):
    document = Document(samples("example.odt"))
    clone = document.clone
    # new xmlparts cache should be empty
    assert clone._Document__xmlparts == {}


def test_case_clone_same_content(samples):
    document = Document(samples("example.odt"))
    s_orig = document.body.serialize()
    clone = document.clone
    s_clone = clone.body.serialize()
    assert s_clone == s_orig


def test_case_clone_different_changes_1(samples):
    document = Document(samples("example.odt"))
    s_orig = document.body.serialize()
    clone = document.clone
    clone.body.append(Paragraph("some text"))
    s_clone = clone.body.serialize()
    assert s_clone != s_orig


def test_case_clone_different_unchanged_1(samples):
    document = Document(samples("example.odt"))
    s_orig = document.body.serialize()
    clone = document.clone
    clone.body.append(Paragraph("some text"))
    s_after = document.body.serialize()
    assert s_after == s_orig


def test_case_clone_different_unchanged_2(samples):
    document = Document(samples("example.odt"))
    clone = document.clone
    clone.body.append(Paragraph("some text"))
    s_clone1 = clone.body.serialize()
    document.body.append(Paragraph("new text"))
    s_clone2 = clone.body.serialize()
    assert s_clone1 == s_clone2


def test_case_save_nogenerator(tmp_path, samples):
    document = Document(samples("example.odt"))
    temp = BytesIO()
    document.save(temp)
    temp.seek(0)
    new = Document(temp)
    generator = new.get_part(ODF_META).get_generator()
    assert generator.startswith("odfdo")


def test_case_save_generator(samples):
    document = Document(samples("example.odt"))
    document.get_part(ODF_META).set_generator("toto")
    temp = BytesIO()
    document.save(temp)
    temp.seek(0)
    new = Document(temp)
    generator = new.get_part(ODF_META).get_generator()
    assert generator == "toto"


def test_repr_empty():
    document = Document()
    assert repr(document) == "<Document type=text path=None>"


def test_repr_text(samples):
    document = Document(samples("example.odt"))
    assert "example.odt" in repr(document)


def test_repr_spreadsheet(samples):
    document = Document(samples("simple_table.ods"))
    result = repr(document)
    assert "type=spreadsheet" in result
    assert "simple_table.ods" in result


def test_str_empty():
    document = Document()
    document.body.clear()
    assert str(document) == ""


def test_str_text(samples):
    document = Document(samples("example.odt"))
    result = str(document)
    assert "odfdo Test Case Document" in result
    assert "This is the second paragraph" in result
    assert "First paragraph[*] of the second section" in result


def test_str_spreadsheet(samples):
    document = Document(samples("simple_table.ods"))
    result = str(document)
    assert "1 1 1 2 3 3 3" in result
    assert "A float" in result
    assert "3.14" in result


def test_document_meta_error():
    doc = Document(None)
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.meta


def test_settings_property():
    doc = Document("text")
    assert doc.settings is not None


def test_document_settings_error():
    doc = Document(None)
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.settings


def test_document_styles_error():
    doc = Document(None)
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.styles


def test_document_manifest_error():
    doc = Document(None)
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.manifest


def test_document_get_type_no_container():
    doc = Document(None)
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.get_type()


def test_mimetype_setter_text_template():
    doc = Document("text")
    doc.mimetype = "application/vnd.oasis.opendocument.text-template"
    assert doc.mimetype == "application/vnd.oasis.opendocument.text-template"


def test_document_mimetype_property_error():
    doc = Document("text")
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.mimetype


def test_document_get_formatted_text_spreadsheet():
    doc = Document("spreadsheet")
    table = Table("T1")
    table.set_value("A1", "cell1")
    doc.body.append(table)
    res = doc.get_formatted_text()
    assert "cell1" in res


def test_document_get_formatted_text_rst():
    doc = Document("text")
    p = Paragraph("test")
    doc.body.append(p)
    res = doc.get_formatted_text(rst_mode=True)
    assert "test" in res


def test_document_get_formatted_text_complex():
    doc = Document("text")
    p = Paragraph("test")

    # Footnote
    fn = Note("footnote", note_id="fn1")
    fn.note_body = Paragraph("note content")
    p.append(fn)

    # Annotation
    ann = Annotation("ann1", creator="me")
    ann.note_body = Paragraph("ann content")
    p.append(ann)

    doc.body.append(p)

    # Endnote
    en = Note("endnote", note_id="en1")
    en.note_body = Paragraph("endnote content")
    p2 = Paragraph("p2")
    p2.append(en)
    doc.body.append(p2)

    # Image for RST mode
    img = DrawImage("Pictures/test.png")
    frame = Frame.image_frame(img, width="10cm", height="10cm")
    doc.body.append(frame)

    res = doc.get_formatted_text()
    assert "test" in res
    assert "note content" in res
    assert "ann content" in res
    assert "endnote content" in res

    res_rst = doc.get_formatted_text(rst_mode=True)
    assert ".. [#]" in res_rst
    assert ".. image::" in res_rst
    assert ":width:" in res_rst
    assert ":height:" in res_rst


def test_document_get_formatted_text_not_implemented():
    doc = Document("text")
    with patch.object(Document, "get_type", return_value="formula"):
        with pytest.raises(NotImplementedError, match="not supported yet"):
            doc.get_formatted_text()


def test_document_del_part_success():
    doc = Document("text")
    doc.container.set_part("some/part", b"data")
    doc.del_part("some/part")
    with pytest.raises(ValueError, match="is deleted"):
        doc.get_part("some/part")


def test_document_del_part_no_container():
    doc = Document("text")
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.del_part("some/part")


def test_document_del_part_mandatory():
    doc = Document("text")
    with pytest.raises(ValueError, match="is mandatory"):
        doc.del_part("content")


def test_document_set_part_no_container():
    doc = Document("text")
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.set_part("some/part", b"data")


def test_document_get_parts_no_container_property():
    doc = Document("text")
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _a = doc.parts


def test_document_new_invalid_template():
    with pytest.raises(FileNotFoundError):
        Document.new("invalid")


def test_document_new_path(tmp_path):
    doc1 = Document("text")
    path = tmp_path / "template.ott"
    doc1.save(path)
    doc2 = Document.new(path)
    assert doc2.container is not None


def test_document_save_no_container():
    doc = Document("text")
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.save("test.odt")


def test_document_save_invalid_packaging():
    doc = Document("text")
    with pytest.raises(ValueError, match="not supported"):
        doc.save("test.odt", packaging="invalid")


def test_document_save_pretty_caching():
    doc = Document("text")
    doc.get_part("content")
    with patch.object(doc.container, "save"):
        doc.save("test.odt", pretty=True)


def test_document_check_manifest_rdf():
    doc = Document("text")
    doc.manifest.set_media_type(ODF_MANIFEST_RDF, "application/rdf+xml")
    doc._check_manifest_rdf()
    assert ODF_MANIFEST_RDF in doc.container.parts

    doc.manifest.del_full_path(ODF_MANIFEST_RDF)
    doc.container.set_part(ODF_MANIFEST_RDF, b"<rdf/>")
    doc._check_manifest_rdf()
    with pytest.raises(ValueError, match="is deleted"):
        doc.get_part(ODF_MANIFEST_RDF)


def test_document_delete_styles():
    doc = Document("text")

    p = Paragraph("test", style="s1")
    doc.body.append(p)
    doc.insert_style(Style("paragraph", name="s1"))

    mock_style = MagicMock(spec=StyleBase)
    type(mock_style).name = PropertyMock(side_effect=AttributeError)
    mock_style2 = MagicMock(spec=StyleBase)
    type(mock_style2).name = PropertyMock(return_value=None)

    with patch.object(Document, "get_styles", return_value=[mock_style, mock_style2]):
        doc.delete_styles()

    deleted = doc.delete_styles()
    assert deleted > 0
    assert p.style is None


def test_document_copy_image_from_document():
    doc1 = Document("text")
    doc2 = Document("text")
    doc1.container.set_part("Pictures/test.png", b"data")
    doc1.manifest.add_full_path("Pictures/test.png", "image/png")
    doc2._copy_image_from_document(doc1, "Pictures/test.png")
    assert doc2.get_part("Pictures/test.png") == b"data"
    assert doc2.manifest.get_media_type("Pictures/test.png") == "image/png"


def test_document_merge_styles_from():
    doc1 = Document("text")
    doc2 = Document("text")
    s1 = Style("paragraph", name="style1")
    doc1.insert_style(s1)
    doc2.merge_styles_from(doc1)
    assert doc2.get_style("paragraph", "style1") is not None


def test_document_merge_styles_from_complex():
    doc1 = Document("text")
    doc2 = Document("text")
    s1 = Style("paragraph", name="auto1")
    doc1.insert_style(s1, automatic=True)
    s2 = StyleMasterPage("master1", page_layout="pl1")
    doc1.insert_style(s2)
    s3 = Style("font-face", font_name="font1")
    doc1.insert_style(s3)

    mock_style = MagicMock(spec=StyleBase)
    type(mock_style).family = PropertyMock(return_value=None)

    mock_style2 = MagicMock(spec=StyleBase)
    type(mock_style2).family = PropertyMock(return_value="paragraph")
    type(mock_style2).parent = PropertyMock(return_value=None)

    with patch.object(Document, "get_styles", return_value=[mock_style, mock_style2]):
        doc2.merge_styles_from(doc1)

    doc2.merge_styles_from(doc1)
    assert doc2.get_style("paragraph", "auto1") is not None
    assert doc2.get_style("master-page", "master1") is not None
    assert doc2.get_style("font-face", "font1") is not None


def test_document_add_page_break_style():
    doc = Document("text")
    doc.add_page_break_style()
    assert doc.get_style("paragraph", "odfdopagebreak") is not None
    doc.add_page_break_style()


def test_document_get_table():
    doc = Document("spreadsheet")
    table = Table("T1")
    doc.body.append(table)
    assert doc._get_table("T1").name == "T1"
    assert doc._get_table(0) is not None
    with pytest.raises(TypeError, match="must be int or str"):
        doc._get_table(None)


def test_document_get_cell_style_properties():
    doc = Document("spreadsheet")
    table = Table("T1")
    table.set_value("A1", "val")
    doc.body.append(table)
    assert doc.get_cell_style_properties("T1", "A1") == {}
    s1 = Style("table-cell", name="s1", background_color="#ff0000")
    doc.insert_style(s1)
    table.get_cell("A1", clone=False).style = "s1"
    props = doc.get_cell_style_properties("T1", "A1")
    assert props.get("fo:background-color") == "#ff0000"


def test_document_get_cell_style_properties_branches():
    doc = Document("spreadsheet")
    table = Table("T1")
    table.set_value("A1", "val")
    doc.body.append(table)

    assert doc.get_cell_style_properties("NonExistent", "A1") == {}

    s_row = Style("table-row", name="sr1")
    s_row.set_properties(area="table-cell", background_color="#0000FF")
    doc.insert_style(s_row)
    table.get_row(0, clone=False).style = "sr1"
    props = doc.get_cell_style_properties("T1", "A1")
    assert props.get("fo:background-color") == "#0000FF"

    table.get_row(0, clone=False).style = None
    s1 = Style("table-cell", name="s1", background_color="#FF0000")
    doc.insert_style(s1)
    col = table.get_column(0)
    col.default_cell_style = "s1"
    table.set_column(0, col)
    props = doc.get_cell_style_properties("T1", "A1")
    assert props.get("fo:background-color") == "#FF0000"

    col = table.get_column(0)
    col.default_cell_style = None
    table.set_column(0, col)
    assert doc.get_cell_style_properties("T1", "A1") == {}

    with patch.object(Table, "get_column", side_effect=ValueError):
        assert doc.get_cell_style_properties("T1", "A1") == {}


def test_document_get_cell_background_color():
    doc = Document("spreadsheet")
    table = Table("T1")
    table.set_value("A1", "val")
    doc.body.append(table)
    assert doc.get_cell_background_color("T1", "A1") == "#ffffff"
    s1 = Style("table-cell", name="s1", background_color="#ff0000")
    doc.insert_style(s1)
    table.get_cell("A1", clone=False).style = "s1"
    assert doc.get_cell_background_color("T1", "A1") == "#ff0000"


def test_document_get_table_style():
    doc = Document("spreadsheet")
    table = Table("T1")
    doc.body.append(table)
    assert doc.get_table_style("NonExistent") is None
    s1 = Style("table", name="st1")
    doc.insert_style(s1)
    table.style = "st1"
    assert doc.get_table_style("T1").name == "st1"


def test_document_get_table_displayed():
    doc = Document("spreadsheet")
    table = Table("T1")
    doc.body.append(table)
    assert doc.get_table_displayed("T1") is True
    doc.set_table_displayed("T1", False)
    assert doc.get_table_displayed("T1") is False

    table.style = None
    assert doc.get_table_displayed("T1") is True


def test_document_unique_style_name():
    doc = Document("text")
    name1 = doc._unique_style_name("base")
    assert name1 == "base_0"
    doc.insert_style(Style("paragraph", name="base_0"))
    name2 = doc._unique_style_name("base")
    assert name2 == "base_1"


def test_document_set_table_displayed():
    doc = Document("spreadsheet")
    table = Table("T1")
    doc.body.append(table)
    doc.set_table_displayed("T1", False)
    style = doc.get_table_style("T1")
    assert style.get_properties()["table:display"] == "false"
    doc.set_table_displayed("T1", True)
    style = doc.get_table_style("T1")  # Get the NEW style
    assert style.get_properties()["table:display"] == "true"


def test_document_language():
    doc = Document("text")
    doc.language = "fr-FR"
    assert doc.language == "fr-FR"
    with pytest.raises(TypeError, match=r"RFC3066"):
        doc.language = "invalid"


def test_document_get_parent_style():
    doc = Document("text")
    s1 = Style("paragraph", name="parent")
    s2 = Style("paragraph", name="child", parent_style="parent")
    doc.insert_style(s1)
    doc.insert_style(s2)
    assert doc.get_parent_style(s2).name == "parent"
    assert doc.get_parent_style(s1) is None


def test_document_get_list_style():
    doc = Document("text")
    ls = Style("list", name="ls1")
    doc.insert_style(ls)
    s1 = Style("paragraph", name="s1")
    s1.set_attribute("style:list-style-name", "ls1")
    doc.insert_style(s1)
    assert doc.get_list_style(s1).name == "ls1"


def test_document_set_automatic_name():
    doc = Document("text")
    s1 = Style("paragraph")
    doc._set_automatic_name(s1, "paragraph")
    assert s1.name.startswith("odfdo_auto_")
    doc.insert_style(s1, automatic=True)
    s2 = Style("paragraph")
    doc._set_automatic_name(s2, "paragraph")
    assert s2.name != s1.name


def test_document_show_styles():
    doc = Document("text")
    s1 = Style("paragraph", name="s1", color="red")
    s1.set_properties(area="paragraph", writing_mode="lr-tb")
    doc.insert_style(s1)
    s2 = DrawFillImage(name="img1")
    doc.insert_style(s2)
    s3 = Style("text", name="auto1")
    doc.insert_style(s3, automatic=True)
    res = doc.show_styles(properties=True)
    assert "s1" in res
    assert "img1" in res
    assert "auto1" in res
    assert "writing-mode" in res


def test_document_pseudo_style_attribute():
    doc = Document("text")
    s1 = Style("paragraph", name="s1", display_name="Display S1")
    res = doc._pseudo_style_attribute(s1, "name")
    assert res == "s1"
    res = doc._pseudo_style_attribute(s1, "display_name")
    assert res == "Display S1"
    res = doc._pseudo_style_attribute(s1, "invalid")
    assert res == ""


def test_document_get_formated_meta():
    doc = Document("text")
    res = doc.get_formated_meta()
    assert "Creation date" in res


def test_document_to_markdown():
    doc = Document("text")
    doc.body.append(Element.from_tag("<text:p>test markdown</text:p>"))
    res = doc.to_markdown()
    assert "test markdown" in res

    doc2 = Document("spreadsheet")
    with pytest.raises(NotImplementedError):
        doc2.to_markdown()


def test_document_add_file(tmp_path):
    doc = Document("text")
    f = tmp_path / "test.txt"
    f.write_text("content")
    path = doc.add_file(f)
    assert "Pictures/" in path

    buf = BytesIO(b"content2")
    buf.name = "test2.txt"
    doc.add_file(buf)


def test_document_clone():
    doc = Document("text")
    # Add a dummy XmlPart to test setattr loop
    doc.dummy_part = XmlPart("dummy", doc.container)
    doc2 = doc.clone
    assert doc2 is not doc
    assert doc2.get_type() == doc.get_type()
    assert hasattr(doc2, "dummy_part")
    assert doc2.dummy_part is not doc.dummy_part


def test_check_manifest_rdf_no_container():
    doc = Document()
    doc.container = None
    # Should return early
    doc._check_manifest_rdf()


def test_show_styles_no_name_style():
    doc = Document("text")
    style_mock = MagicMock(spec=Style)
    style_mock.tag = "style:style"
    style_mock.name = ""  # instead of None to avoid MagicMock in join
    style_mock.family = "paragraph"
    style_mock.parent = None  # is_auto False
    style_mock.attributes = {}
    style_mock.children = []
    # Mocking for _pseudo_style_attribute
    style_mock.display_name = ""
    style_mock.parent_style = ""

    with patch.object(Document, "get_styles", return_value=[style_mock]):
        res = doc.show_styles()
        assert "family:paragraph" in res


def test_delete_styles_name_none():
    doc = Document("text")
    style_mock = MagicMock(spec=Style)
    style_mock.name = None
    with patch.object(Document, "get_styles", return_value=[style_mock]):
        # Should skip style with name None (now via if not name:)
        assert doc.delete_styles() == 0


def test_get_parts_no_container():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.get_parts()


def test_get_type_invalid_mimetype():
    doc = Document()
    doc.container = MagicMock()
    # Mimetype without dot
    doc.container.mimetype = "text"
    assert doc.get_type() == "text"


def test_content_property_none():
    doc = Document("text")
    with patch.object(Document, "get_part", return_value=None):
        with pytest.raises(ValueError, match="Empty Content"):
            _ = doc.content


def test_meta_property_none():
    doc = Document("text")
    with patch.object(Document, "get_part", return_value=None):
        with pytest.raises(ValueError, match="Empty Meta"):
            _ = doc.meta


def test_styles_property_none():
    doc = Document("text")
    with patch.object(Document, "get_part", return_value=None):
        with pytest.raises(ValueError, match="Empty Styles"):
            _ = doc.styles


def test_get_parent_style_none():
    doc = Document("text")
    assert doc.get_parent_style(None) is None


def test_get_list_style_none():
    doc = Document("text")
    assert doc.get_list_style(None) is None


def test_get_list_style_no_attr():
    doc = Document("text")
    style = Element.from_tag("style:style")
    assert doc.get_list_style(style) is None


def test_get_list_style_no_name_branch():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    style.set_attribute("style:list-style-name", "")
    assert doc.get_list_style(style) is None


def test_show_styles_drawfillimage_branch():
    doc = Document("text")
    fill = DrawFillImage("myfill", "Pictures/img.png")
    # DrawFillImage is usually inserted in office:styles (common)
    doc.insert_style(fill)
    res = doc.show_styles()
    assert "myfill" in res
    # The family for DrawFillImage should be "" in show_styles
    assert "family: " in res


def test_copy_image_from_document_not_bytes():
    doc1 = Document("text")
    doc2 = Document("text")
    with patch.object(Document, "get_part", return_value="not bytes"):
        doc2._copy_image_from_document(doc1, "path")
        assert "path" not in doc2.parts


def test_merge_styles_from_no_name_attr():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Element.from_tag("style:style")
    style.set_attribute("style:family", "paragraph")
    doc1.content.get_element("//office:automatic-styles").append(style)
    doc2.merge_styles_from(doc1)


def test_merge_styles_from_stylename_none():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph")
    doc1.content.get_element("//office:automatic-styles").append(style)
    doc2.merge_styles_from(doc1)


def test_get_style_properties_style_none():
    doc = Document("text")
    assert doc.get_style_properties("paragraph", "NONEXISTENT") is None


def test_get_cell_style_properties_row_style_none():
    doc = Document("spreadsheet")
    sheet = doc.body.get_table(0)
    row = sheet.get_row(0)
    row.style = None
    props = doc.get_cell_style_properties(0, (0, 0))
    assert isinstance(props, dict)


def test_get_cell_style_properties_column_style_none():
    doc = Document("spreadsheet")
    sheet = doc.body.get_table(0)
    col = sheet.get_column(0)
    col.default_cell_style = None
    props = doc.get_cell_style_properties(0, (0, 0))
    assert props == {}


def test_mimetype_empty_container_old():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.mimetype
    with pytest.raises(ValueError, match="Empty Container"):
        doc.mimetype = "application/vnd.oasis.opendocument.text"


def test_meta_empty_or_invalid():
    doc = Document("text")
    doc.get_part = MagicMock(return_value=None)
    with pytest.raises(ValueError, match="Empty Meta"):
        _ = doc.meta


def test_manifest_empty_or_invalid():
    doc = Document("text")
    doc.get_part = MagicMock(return_value=None)
    with pytest.raises(ValueError, match="Empty Manifest"):
        _ = doc.manifest


def test_settings_empty_or_invalid():
    doc = Document("text")
    doc.get_part = MagicMock(return_value=None)
    with pytest.raises(ValueError, match="Empty settings part"):
        _ = doc.settings


def test_to_markdown_invalid_type():
    doc = Document("spreadsheet")
    with pytest.raises(
        NotImplementedError, match="Type of document 'spreadsheet' not supported yet"
    ):
        doc.to_markdown()


def test_add_binary_part_empty_container_old():
    doc = Document()
    doc.container = None
    blob = Blob.from_io(BytesIO(b"content"))
    with pytest.raises(ValueError, match="Empty Container"):
        doc._add_binary_part(blob)


def test_add_file_empty_container_old():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.add_file(BytesIO(b"content"))


def test_clone_empty_container_old():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.clone


def test_get_parent_style_no_family():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    style.family = None
    assert doc.get_parent_style(style) is None


def test_get_parent_style_no_parent():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    assert doc.get_parent_style(style) is None


def test_get_list_style_no_name():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    assert doc.get_list_style(style) is None


def test_set_automatic_name_no_name_attr_old():
    doc = Document("text")
    element = Element.from_tag("dc:title")
    doc._set_automatic_name(element, "paragraph")
    assert not hasattr(element, "name")


def test_show_styles_flags_old():
    doc = Document("text")
    res = doc.show_styles(automatic=False)
    assert "auto" not in res
    res = doc.show_styles(common=False)
    assert "common" not in res
    res = doc.show_styles(automatic=False, common=False)
    assert res == ""


def test_delete_styles_mixed_old():
    doc = Document("text")
    style1 = Style("paragraph", name="USED")
    doc.insert_style(style1, automatic=False)
    p = Paragraph("test", style="USED")
    doc.body.append(p)
    count = doc.delete_styles()
    assert count >= 1
    assert p.get_attribute("text:style-name") is None
    assert doc.get_style("paragraph", "USED") is None


def test_add_page_break_style_existing_old():
    doc = Document("text")
    doc.add_page_break_style()
    doc.add_page_break_style()
    assert doc.get_style("paragraph", "odfdopagebreak") is not None


def test_get_table_invalid_type_old():
    doc = Document("spreadsheet")
    with pytest.raises(TypeError, match="Table parameter must be int or str"):
        doc._get_table(None)


def test_save_pretty_xml_old_old():
    doc = Document("text")
    out = BytesIO()
    doc.save(out, packaging="xml", pretty=True)
    assert b"<office:document" in out.getvalue()


def test_check_manifest_rdf_missing_part_old_old():
    doc = Document("text")
    doc.manifest.add_full_path("manifest.rdf", "application/rdf+xml")
    doc._check_manifest_rdf()
    assert "manifest.rdf" in doc.container.parts


def test_merge_styles_from_different_containers_old():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    doc1.content.get_element("//office:automatic-styles").append(style)
    doc2.merge_styles_from(doc1)
    assert doc2.content.get_style("paragraph", "P1") is not None


def test_merge_styles_from_master_page_with_image_old():
    doc1 = Document("text")
    doc2 = Document("text")
    master_page = Element.from_tag("style:master-page")
    master_page.set_attribute("style:name", "MyMaster")
    doc1.styles.get_element("//office:master-styles").append(master_page)
    img_path = "Pictures/test.png"
    img_content = b"fake image"
    doc1.set_part(img_path, img_content)
    doc1.manifest.add_full_path(img_path, "image/png")
    img = DrawImage(url=img_path)
    master_page.append(img)
    doc2.merge_styles_from(doc1)
    assert doc2.get_part(img_path) == img_content


def test_merge_styles_from_draw_fill_image_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    img_path = "Pictures/test_fill.png"
    img_content = b"fake fill image"
    doc1.set_part(img_path, img_content)
    doc1.manifest.add_full_path(img_path, "image/png")
    fill_image = DrawFillImage(name="MyFill", url=img_path)
    doc1.styles.get_element("//office:styles").append(fill_image)
    doc2.merge_styles_from(doc1)
    assert doc2.get_part(img_path) == img_content


def test_insert_style_font_face_decls_old_old():
    doc = Document("text")
    font_face = Element.from_tag("style:font-face")
    font_face.set_attribute("style:name", "MyFont")
    doc.content.get_element("//office:font-face-decls").append(font_face)
    doc.insert_style(font_face)
    assert doc.get_style("font-face", "MyFont") is not None


def test_merge_styles_from_content_automatic_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    doc1.content.get_element("//office:automatic-styles").append(style)
    doc2.merge_styles_from(doc1)
    assert doc2.content.get_style("paragraph", "P1") is not None


def test_get_formatted_text_images_rst_old_old():
    doc = Document("text")
    result = []
    context = {
        "images": [
            ("ref1", "file1.png", ("10cm", "5cm")),
            ("ref2", "file2.png", (None, None)),
        ],
        "annotations": [],
        "endnotes": [],
    }
    doc._get_formatted_text_images(result, context, rst_mode=True)
    res_str = "".join(result)
    assert ".. ref1 image:: file1.png" in res_str
    assert ":width: 10cm" in res_str
    assert ":height: 5cm" in res_str
    assert ".. ref2 image:: file2.png" in res_str
    assert "width" not in res_str.split("file2.png")[-1]


def test_set_automatic_name_index_parsing_old_old():
    doc = Document("text")
    style1 = Style("paragraph", name="odfdo_auto_abc")
    doc.content.get_element("//office:automatic-styles").append(style1)
    style2 = Style("paragraph", name="NewStyle")
    doc._set_automatic_name(style2, "paragraph")
    assert style2.name == "odfdo_auto_1"


def test_show_styles_with_properties_old_old():
    doc = Document("text")
    style = Style("paragraph", name="P_PROPS")
    style.set_properties(area="paragraph", align="center")
    doc.styles.get_element("//office:styles").append(style)
    p = Paragraph("test", style="P_PROPS")
    doc.body.append(p)
    res = doc.show_styles(properties=True)
    assert "P_PROPS" in res
    assert "fo:text-align: center" in res


def test_merge_styles_from_duplicate_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    style1 = Style("paragraph", name="P_DUP")
    style1.set_properties(area="text", color="#111111")
    doc1.styles.get_element("//office:styles").append(style1)
    style2 = Style("paragraph", name="P_DUP")
    style2.set_properties(area="text", color="#222222")
    doc2.styles.get_element("//office:styles").append(style2)
    doc2.merge_styles_from(doc1)
    res_style = doc2.get_style("paragraph", "P_DUP")
    assert res_style.get_properties(area="text")["fo:color"] == "#111111"


def test_get_parent_style_not_found_old_old():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle", parent_style="NonExistent")
    assert doc.get_parent_style(style) is None


def test_get_list_style_not_found_old_old():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    style.list_style_name = "NonExistent"
    assert doc.get_list_style(style) is None


def test_get_styles_family_bytes_old_old():
    doc = Document("text")
    styles = doc.get_styles(family=b"paragraph")
    assert len(styles) > 0


def test_get_styled_elements_all_old_old():
    doc = Document("text")
    elements = doc.get_styled_elements()
    assert len(elements) > 0


def test_delete_styles_attribute_error_old_old():
    doc = Document("text")
    style = MagicMock()
    type(style).name = property(
        lambda x: exec('raise AttributeError("no name")'),  # noqa: S102
    )
    orig_get_styles = doc.get_styles
    doc.get_styles = MagicMock(return_value=orig_get_styles() + [style])
    res = doc.delete_styles()
    assert isinstance(res, int)


def test_copy_image_from_document_invalid_content_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    doc1.get_part = MagicMock(return_value=None)
    doc2._copy_image_from_document(doc1, "Pictures/none.png")
    assert "Pictures/none.png" not in doc2.container.parts


def test_merge_styles_from_no_name_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Element.from_tag("style:style")
    style.set_attribute("style:family", "paragraph")
    doc1.styles.get_element("//office:styles").append(style)
    doc2.merge_styles_from(doc1)
    assert doc2.styles.get_style("paragraph", None) is not None


def test_merge_styles_from_wrong_upper_container_old_old():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P_WRONG")
    wrong_container = Element.from_tag("office:body")
    wrong_container.append(style)
    doc1.get_styles = MagicMock(return_value=[style])
    with pytest.raises(NotImplementedError):
        doc2.merge_styles_from(doc1)


def test_save_pretty_zip_old_old():
    doc = Document("text")
    out = BytesIO()
    doc.save(out, packaging="zip", pretty=True)
    from zipfile import ZipFile

    with ZipFile(out) as z:
        content = z.read("content.xml")
        assert b"\n" in content


def test_get_parts_empty_container_old_old():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc.get_parts()


def test_get_type_all_old_old():
    for mimetype, expected in [
        ("application/vnd.oasis.opendocument.text", "text"),
        ("application/vnd.oasis.opendocument.spreadsheet", "spreadsheet"),
    ]:
        doc = Document()
        doc.container = MagicMock()
        doc.container.mimetype = mimetype
        assert doc.get_type() == expected


def test_get_formatted_text_footnotes_non_rst_old_old():
    doc = Document("text")
    result = []
    context = {
        "footnotes": [("1", "note 1")],
        "annotations": [],
        "images": [],
        "endnotes": [],
    }
    doc._get_formatted_text_footnotes(result, context, rst_mode=False)
    assert "note 1" in "".join(result)


def test_get_formatted_text_footnotes_non_rst():
    doc = Document("text")
    result = []
    context = {
        "footnotes": [("1", "citation")],
        "annotations": [],
        "images": [],
        "endnotes": [],
    }
    doc._get_formatted_text_footnotes(result, context, rst_mode=False)
    assert "----\n" in "".join(result)
    assert "[1] citation\n" in "".join(result)


def test_get_formatted_text_images_call_via_header():
    doc = Document("text")
    header = Header(1, "Title")
    img = DrawImage(url="Pictures/test.png")
    frame = Frame.image_frame(img, size=("1cm", "1cm"))
    header.append(frame)
    doc.body.append(header)
    res = doc.get_formatted_text(rst_mode=True)
    assert ".. |img1| image:: Pictures/test.png" in res


def test_check_manifest_rdf_extra_in_parts_old_old():
    doc = Document("text")
    doc.manifest.get_media_type = MagicMock(return_value=None)
    doc.container.get_parts = MagicMock(return_value=[ODF_MANIFEST_RDF, "mimetype"])
    doc.container.del_part = MagicMock()
    doc._check_manifest_rdf()
    doc.container.del_part.assert_called_with(ODF_MANIFEST_RDF)


def test_get_style_with_style_base_old_old():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    assert doc.get_style("paragraph", style) is style


def test_get_style_default_old_old():
    doc = Document("text")
    style = doc.get_style("paragraph", None)
    assert style is not None
    assert style.tag == "style:default-style"


def test_set_automatic_name_with_existing_indices_old_old():
    doc = Document("text")
    style1 = Style("paragraph", name="odfdo_auto_10")
    doc.content.get_element("//office:automatic-styles").append(style1)
    style2 = Style("paragraph", name="NewStyle")
    doc._set_automatic_name(style2, "paragraph")
    assert style2.name == "odfdo_auto_11"


def test_delete_styles_with_attributes_old_old():
    doc = Document("text")
    p = Paragraph("test", style="Standard")
    doc.body.append(p)
    doc.delete_styles()
    assert p.get_attribute("text:style-name") is None


def test_get_part_none_old_old():
    doc = Document("text")
    doc.container.get_part = MagicMock(return_value=None)
    res = doc.get_part("non_existent_raw_part")
    assert res is None


def test_get_table_style_old_old():
    doc = Document("spreadsheet")
    style = doc.get_table_style(0)
    assert style is not None
    assert style.family == "table"


def test_get_table_displayed_old_old():
    doc = Document("spreadsheet")
    assert doc.get_table_displayed(0) is True


def test_set_table_displayed_old_old():
    doc = Document("spreadsheet")
    doc.set_table_displayed(0, False)
    assert doc.get_table_displayed(0) is False


def test_get_cell_style_properties_cell_old_old():
    doc = Document("spreadsheet")
    sheet = doc.body.get_table(0)
    style = Style("table-cell", name="cellstyle")
    style.set_properties(area="table-cell", background_color="#112233")
    doc.insert_style(style, automatic=False)
    sheet.set_value((0, 0), value="", style="cellstyle")
    res_props = doc.get_cell_style_properties(0, (0, 0))
    assert res_props.get("fo:background-color") == "#112233"


def test_get_cell_style_properties_column_old_old():
    doc = Document("spreadsheet")
    sheet = doc.body.get_table(0)
    column = sheet.get_column(0)
    column.default_cell_style = "colstyle"
    sheet.set_column(0, column)
    style = Style("table-cell", name="colstyle")
    style.set_properties(area="table-cell", background_color="#654321")
    doc.insert_style(style, automatic=False)
    cell = sheet.get_cell((0, 0))
    cell.style = None
    sheet.set_cell((0, 0), cell)
    row = sheet.get_row(0)
    row.style = None
    sheet.set_row(0, row)
    res_props = doc.get_cell_style_properties(0, (0, 0))
    assert res_props.get("fo:background-color") == "#654321"


def test_get_set_language_old_old():
    doc = Document("text")
    doc.language = "fr-FR"
    assert doc.language == "fr-FR"
    assert doc.meta.language == "fr-FR"
    with pytest.raises(TypeError, match="Language must be"):
        doc.language = "invalid"


def test_save_pretty_xml_runtime_error():
    doc = Document("text")
    _ = doc.meta

    def mock_get_part_class(path):
        if path == "settings.xml":
            return None
        return _get_part_class(path)

    with patch("odfdo.document._get_part_class", side_effect=mock_get_part_class):
        out = BytesIO()
        with pytest.raises(RuntimeError, match="Should never happen"):
            doc.save(out, pretty=True)


def test_save_pretty_none_xmlparts_1():
    doc = Document("text")
    doc._Document__xmlparts[ODF_CONTENT] = None
    out = BytesIO()
    doc.save(out, pretty=True)
    assert out.getvalue() != b""


def test_save_not_pretty_none_xmlparts_2():
    doc = Document("text")
    doc._Document__xmlparts[ODF_CONTENT] = None
    out = BytesIO()
    doc.save(out, pretty=False)
    assert out.getvalue() != b""


def test_save_xmlparts_path_known():
    doc = Document("text")
    _ = doc.content
    _ = doc.meta
    _ = doc.styles
    _ = doc.settings
    out = BytesIO()
    doc.save(out, pretty=True)
    assert out.getvalue() != b""


def test_save_container_get_part_none():
    doc = Document("text")
    valid_meta = (
        b"<office:document-meta xmlns:office="
        b'"urn:oasis:names:tc:opendocument:xmlns:office:1.0" '
        b'xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0" '
        b'office:version="1.2">'
        b"<office:meta>"
        b"<meta:generator/>"
        b"</office:meta>"
        b"</office:document-meta>"
    )

    def side_effect(path):
        if path == "settings.xml":
            return None
        if path == "meta.xml":
            return valid_meta
        return b"<xml/>"

    with patch.object(doc.container, "get_part", side_effect=side_effect):
        out = BytesIO()
        doc.save(out, pretty=True)
    assert out.getvalue() != b""



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from datetime import datetime
from unittest.mock import MagicMock, PropertyMock, patch

import pytest

from odfdo.annotation import Annotation, AnnotationEnd
from odfdo.element import Element, EText
from odfdo.mixin_paragraph import (
    _by_offset_wrapper,
    _by_regex_offset,
    _by_regex_wrapper,
)
from odfdo.note import Note
from odfdo.paragraph import Paragraph, Span
from odfdo.reference import (
    ReferenceMark,
    ReferenceMarkEnd,
    ReferenceMarkStart,
)
from odfdo.spacer import Spacer


def _make_span_method(element, *args, **kwargs):
    return Element.from_tag("text:span")


def test_by_offset_wrapper_counted():
    elem = Paragraph("abc")
    span = Span("def")
    elem.append(span)

    res = _by_offset_wrapper(_make_span_method, elem, offset=4, length=1)
    assert len(res) == 1


def test_by_offset_wrapper_no_container_parent():
    elem = Paragraph("abc")

    with patch.object(EText, "parent", new_callable=PropertyMock) as mock_parent:
        mock_parent.return_value = None
        res = _by_offset_wrapper(_make_span_method, elem, offset=0)
        assert res == []


def test_by_offset_wrapper_length_0():
    elem = Paragraph("abc")

    res = _by_offset_wrapper(_make_span_method, elem, offset=1, length=0)
    assert len(res) == 1


def test_by_regex_wrapper_no_regex():
    res = _by_regex_wrapper(lambda: None, Paragraph("test"), "")
    assert res == []


def test_by_regex_offset_wrapper_regex():
    @_by_regex_offset
    def my_method(element, *args, **kwargs):
        return Element.from_tag("text:span")

    elem = Paragraph("hello")
    res = my_method(elem, regex="ll")
    assert len(res) == 1


def test_expand_spaces_not_s():
    elem = Paragraph()
    elem.append("a")
    span = Element.from_tag("text:span")
    span.text = "b"
    elem.append(span)
    res = elem._expand_spaces("c")
    assert len(res) == 3


def test_sub_merge_spaces_complex():
    res = Paragraph._sub_merge_spaces("  a  b  ")
    assert len(res) >= 5


def test_append_plain_text_variants():
    elem = Paragraph()
    elem.append_plain_text(None)
    assert elem.inner_text == ""
    elem.append_plain_text(b"hello")
    assert elem.inner_text == "hello"


def test_append_element_special_tags():
    elem = Paragraph()
    s = Element.from_tag("text:s")
    s.set_attribute("text:c", "2")
    elem.append(s)
    assert elem.inner_text == "  "


def test_append_not_formatted():
    elem = Paragraph()
    elem.append("  a  \n  b  ", formatted=False)
    assert "a b" in elem.get_formatted_text()


def test_insert_note_existing():
    elem = Paragraph("test")
    note = Note("footnote", note_id="id0", citation="*")
    elem.insert_note(
        note,
        note_class="endnote",
        note_id="id1",
        citation="**",
        body="body",
    )
    assert note.note_class == "endnote"
    assert note.note_id == "id1"
    assert note.citation == "**"


def test_insert_note_after_element():
    elem = Paragraph("abc")
    span = Span("s")
    elem.append(span)
    elem.insert_note(after=span, note_id="id1", citation="*")
    assert span.xpath("descendant::text:note")


def test_insert_annotation_existing():
    elem = Paragraph("test")
    ann = Annotation(body="initial", creator="me")
    dt = datetime(2023, 1, 1)
    elem.insert_annotation(
        ann,
        body="new body",
        creator="you",
        date=dt,
    )
    assert ann.note_body == "new body"
    assert ann.creator == "you"
    assert ann.date == dt


def test_insert_annotation_content_element_empty():
    elem = Paragraph()
    inner = Span("")
    elem.append(inner)
    elem.insert_annotation(content=inner, body="ann", creator="me")
    assert inner.xpath("following-sibling::office:annotation")


def test_insert_annotation_position_tuple():
    elem = Paragraph("hello world")
    elem.insert_annotation(position=(1, 5), body="ann", creator="me")
    assert len(elem.xpath("//office:annotation")) == 1
    assert len(elem.xpath("//office:annotation-end")) == 1


def test_insert_annotation_bad_args():
    elem = Paragraph("test")
    with pytest.raises(ValueError, match="Annotation must have a creator"):
        elem.insert_annotation(content="regex", position=(1, 2), body="ann")


def test_insert_annotation_end_invalid():
    elem = Paragraph("test")
    with pytest.raises(ValueError):
        elem.insert_annotation_end(None)
    with pytest.raises(TypeError):
        elem.insert_annotation_end(Paragraph())


def test_set_reference_mark_content_element_empty():
    elem = Paragraph()
    inner = Span("")
    elem.append(inner)
    elem.set_reference_mark("ref1", content=inner)
    assert inner.xpath("following-sibling::text:reference-mark")


def test_set_reference_mark_position_tuple():
    elem = Paragraph("hello")
    elem.set_reference_mark("ref1", position=(1, 3))
    assert len(elem.xpath("//text:reference-mark-start")) == 1
    assert len(elem.xpath("//text:reference-mark-end")) == 1


def test_set_reference_mark_bad_args():
    elem = Paragraph("test")
    with pytest.raises(ValueError, match="bad arguments"):
        elem.set_reference_mark("ref1", content="regex", position=(1, 2))


def test_set_reference_mark_end_from_mark():
    elem = Paragraph("abc")
    mark = ReferenceMark("ref1")
    elem.append(mark)
    elem.set_reference_mark_end(mark, position=1)
    assert mark.tag == "text:reference-mark-start"
    assert len(elem.xpath("//text:reference-mark-end")) == 1


def test_remove_spans_no_keep():
    elem = Paragraph()
    span = Span("text")
    elem.append(span)
    res = elem.remove_spans(keep_heading=False)
    assert len(res.xpath("//text:span")) == 0


def test_insert_reference_no_display_no_body():
    elem = Paragraph("test")
    elem.insert_reference("ref1", ref_format="text")
    ref = elem.get_element("//text:reference-ref")
    assert ref.text == " "


def test_set_bookmark_content_int_position():
    elem = Paragraph("abc")
    elem.text = "abc"
    elem.set_bookmark("bm1", content="b", position=0)
    assert len(elem.xpath("//text:bookmark-start")) == 1
    assert len(elem.xpath("//text:bookmark-end")) == 1


def test_set_bookmark_position_tuple():
    elem = Paragraph("abc")
    elem.text = "abc"
    elem.set_bookmark("bm1", position=(1, 2))
    assert len(elem.xpath("//text:bookmark-start")) == 1
    assert len(elem.xpath("//text:bookmark-end")) == 1


def test_set_bookmark_role_start():
    elem = Paragraph("abc")
    elem.text = "abc"
    elem.set_bookmark("bm1", role="start", position=1)
    assert len(elem.xpath("//text:bookmark-start")) == 1


def test_set_bookmark_role_end():
    elem = Paragraph("abc")
    elem.text = "abc"
    elem.set_bookmark("bm1", role="end", position=1)
    assert len(elem.xpath("//text:bookmark-end")) == 1


def test_set_bookmark_bad_args():
    elem = Paragraph("test")
    with pytest.raises(ValueError, match="bad arguments"):
        elem.set_bookmark("bm1", content="regex", position=(1, 2))
    with pytest.raises(ValueError, match="bad arguments"):
        elem.set_bookmark("bm1", role="invalid")


def test_by_regex_wrapper_container_no_parent():
    elem = Paragraph("abc")

    def method(element, *args, **kwargs):
        return Element.from_tag("text:span")

    with patch.object(
        EText,
        "parent",
        new_callable=PropertyMock,
    ) as mock_parent:
        mock_parent.return_value = None
        res = _by_regex_wrapper(method, elem, regex="a")
        assert res == []


def test_by_offset_wrapper_upper_tail():
    para = Paragraph("a")
    span = Span("b")
    para.append(span)
    span.tail = "c"
    # offset 2 is in tail of span
    res = _by_offset_wrapper(_make_span_method, para, offset=2, length=1)
    assert len(res) == 1


def test_by_regex_wrapper_upper_tail():
    para = Paragraph("a")
    span = Span("b")
    para.append(span)
    span.tail = "c"
    res = _by_regex_wrapper(_make_span_method, para, regex="c")
    assert len(res) == 1


def test_expand_spaces_s_tag():
    elem = Paragraph()
    s = Element.from_tag("text:s")
    s.text = "  "
    elem.append(s)
    res = elem._expand_spaces("x")
    assert "  x" in "".join([str(x) for x in res])


def test_sub_merge_spaces_only_spaces():
    res = Paragraph._sub_merge_spaces("   ")
    assert len(res) == 1
    assert isinstance(res[0], Spacer)


def test_replace_tabs_lb():
    elem = Paragraph()
    res = elem._replace_tabs_lb(["a\tb\nc", Element.from_tag("text:s")])
    assert len(res) == 6


def test_insert_annotation_content_element_not_empty():
    elem = Paragraph()
    inner = Span("text")
    elem.append(inner)
    elem.insert_annotation(content=inner, body="ann", creator="me")
    assert inner.children[0].tag == "office:annotation"
    assert inner.children[-1].tag == "office:annotation-end"


def test_insert_annotation_end_exists():
    elem = Paragraph("abc")
    ann = Annotation(body="ann", creator="me", name="ann1")
    elem.append(ann)
    end1 = AnnotationEnd(ann)
    elem.append(end1)
    elem.insert_annotation_end(ann, position=1)
    assert len(elem.xpath("//office:annotation-end")) == 1


def test_set_reference_mark_content_element_not_empty():
    elem = Paragraph()
    inner = Span("text")
    elem.append(inner)
    elem.set_reference_mark("ref1", content=inner)
    assert inner.children[0].tag == "text:reference-mark-start"
    assert inner.children[-1].tag == "text:reference-mark-end"


def test_set_reference_mark_end_exists():
    elem = Paragraph("abc")
    mark = ReferenceMarkStart("ref1")
    elem.append(mark)
    end1 = ReferenceMarkEnd("ref1")
    elem.append(end1)
    elem.set_reference_mark_end(mark, position=1)
    assert len(elem.xpath("//text:reference-mark-end")) == 1


def test_insert_variable():
    elem = Paragraph("abc")
    var = Element.from_tag("text:variable-set")
    elem.insert_variable(var, after="a")
    assert elem.xpath("//text:variable-set")


def test_remove_span_list():
    elem = Paragraph("abc")
    span1 = elem.set_span("s1", regex="a")[0]
    span2 = elem.set_span("s2", regex="b")[0]
    res = elem.remove_span([span1, span2])
    assert len(res.xpath("//text:span")) == 0


def test_remove_links():
    elem = Paragraph("abc")
    elem.set_link("http://a", regex="a")
    res = elem.remove_links()
    assert len(res.xpath("//text:a")) == 0


def test_remove_link():
    elem = Paragraph("abc")
    link = elem.set_link("http://a", regex="a")[0]
    res = elem.remove_link(link)
    assert len(res.xpath("//text:a")) == 0


def test_insert_reference_with_mark():
    mark = ReferenceMarkStart("ref1")
    mark.set_attribute("text:name", "ref1")
    with patch.object(
        ReferenceMarkStart, "referenced_text", return_value="referenced text"
    ):
        doc = MagicMock()
        doc.get_reference_mark.return_value = mark
        para = Paragraph("ref here: ")
        with patch.object(
            Element, "document_body", new_callable=PropertyMock
        ) as mock_body:
            mock_body.return_value = doc
            para.insert_reference("ref1", ref_format="text")
    ref = para.get_element("//text:reference-ref")
    assert ref.text == "referenced text"


def test_by_offset_wrapper_tail_no_upper():
    para = Paragraph("abc")
    mock_text = MagicMock(spec=EText)
    mock_text.is_text.return_value = False
    mock_text.__len__.return_value = 3
    container = MagicMock(spec=Element)
    container.parent = None
    container.tail = "tail"
    mock_text.parent = container

    with patch.object(Element, "xpath", return_value=[mock_text]):
        res = _by_offset_wrapper(_make_span_method, para, offset=0, length=1)
        assert len(res) == 1


def test_by_regex_wrapper_tail_no_upper():
    para = Paragraph("abc")
    mock_text = MagicMock(spec=EText)
    mock_text.is_text.return_value = False
    mock_text.__str__.return_value = "tail"
    container = MagicMock(spec=Element)
    container.parent = None
    container.tail = "tail"
    mock_text.parent = container

    with patch.object(Element, "xpath", return_value=[mock_text]):
        res = _by_regex_wrapper(_make_span_method, para, regex="tail")
        assert len(res) == 1


def test_by_regex_offset_with_offset():
    @_by_regex_offset
    def method(element, *args, **kwargs):
        return Element.from_tag("text:span")

    para = Paragraph("abc")
    res = method(para, offset=1, length=1)
    assert len(res) == 1


def test_merge_spaces_with_element():
    para = Paragraph()
    span = Span("test")
    res = para._merge_spaces([span])
    assert res == [span]


def test_sub_merge_spaces_with_non_str():
    para = Paragraph()
    span = Span("test")
    # Patch the whole pattern object
    with patch("odfdo.mixin_paragraph._re_spaces_split") as mock_re:
        mock_re.split.return_value = [" ", span, " "]
        res = para._sub_merge_spaces("anything")
        assert span in res
        assert any(isinstance(x, Spacer) for x in res)

    with patch("odfdo.mixin_paragraph._re_spaces_split") as mock_re:
        mock_re.split.return_value = ["a", span]
        res = para._sub_merge_spaces("anything")
        assert res[1] == span


def test_replace_tabs_lb_empty_text():
    para = Paragraph()
    assert para._sub_replace_tabs_lb("") == []


def test_replace_tabs_lb_text_continue():
    para = Paragraph()
    res = para._sub_replace_tabs_lb("a\n\nb")
    assert len(res) == 4


def test_unformatted_variants():
    para = Paragraph()
    assert para._unformatted("") == ""
    assert para._unformatted(b"abc") == "abc"


def test_insert_note_existing_modify():
    para = Paragraph("test")
    # Use orphan Note
    note = Note("footnote", note_id="id1", citation="1", body="body1")
    # All True
    para.insert_note(
        note,
        note_class="endnote",
        note_id="id2",
        citation="2",
        body="body2",
    )
    assert note.note_class == "endnote"

    # All False
    note2 = Note("footnote", note_id="id3", citation="3", body="body3")
    para.insert_note(
        note2,
        note_class=None,
        note_id=None,
        citation=None,
        body=None,
    )


def test_insert_note_after_str():
    para = Paragraph("abc")
    para.insert_note(after="b", note_id="id1", citation="1", body="body")
    assert len(para.xpath("//text:note")) == 1


def test_insert_annotation_existing_modify():
    para = Paragraph("test")
    # Use orphan Annotation
    ann = Annotation("b1", creator="c1")
    dt = datetime(2023, 1, 1)
    # All True
    para.insert_annotation(ann, body="b2", creator="c2", date=dt)
    assert ann.note_body == "b2"

    # All False
    ann2 = Annotation("b3", creator="c3")
    para.insert_annotation(ann2, body=None, creator=None, date=None)


def test_insert_annotation_after_element():
    para = Paragraph("abc")
    span = Span("s")
    para.append(span)
    para.insert_annotation(after=span, body="ann", creator="me")
    assert span.xpath("office:annotation")


def test_insert_annotation_content_regex():
    para = Paragraph("hello world")
    para.insert_annotation(content="world", body="ann", creator="me")
    assert len(para.xpath("//office:annotation")) == 1
    assert len(para.xpath("//office:annotation-end")) == 1


def test_insert_annotation_bad_args_coverage():
    para = Paragraph("abc")
    with pytest.raises(ValueError, match="Bad arguments"):
        para.insert_annotation(
            content="a",
            position=(0, 1),
            body="ann",
            creator="me",
        )


def test_insert_annotation_end_replace():
    para = Paragraph("abc")
    ann = Annotation("ann", creator="me", name="ann1")
    para.append(ann)
    # Mock get_annotation_end to return an element
    end1 = AnnotationEnd(ann)
    # Force the True branch of line 649
    with patch.object(Paragraph, "get_annotation_end", return_value=end1):
        para.insert_annotation_end(ann, position=1)
    # Force the False branch of line 649
    with patch.object(Paragraph, "get_annotation_end", return_value=None):
        para.insert_annotation_end(ann, position=1)


def test_set_reference_mark_content_regex():
    para = Paragraph("hello world")
    para.set_reference_mark("ref1", content="world")
    assert len(para.xpath("//text:reference-mark-start")) == 1
    assert len(para.xpath("//text:reference-mark-end")) == 1


def test_set_reference_mark_positional():
    para = Paragraph("abc")
    para.set_reference_mark("ref1", position=1)
    assert len(para.xpath("//text:reference-mark")) == 1


def test_set_reference_mark_end_type_error():
    para = Paragraph("abc")
    with pytest.raises(TypeError, match="Not a ReferenceMark"):
        para.set_reference_mark_end(Element.from_tag("text:p"))


def test_set_reference_mark_end_replace():
    para = Paragraph("abc")
    mark = ReferenceMarkStart("ref1")
    para.append(mark)
    end1 = ReferenceMarkEnd("ref1")
    with patch.object(
        Paragraph,
        "get_reference_mark_end",
        return_value=end1,
    ):
        para.set_reference_mark_end(mark, position=1)
    with patch.object(
        Paragraph,
        "get_reference_mark_end",
        return_value=None,
    ):
        para.set_reference_mark_end(mark, position=1)


def test_remove_spans_variants():
    para = Paragraph("abc")
    para.remove_spans(keep_heading=False)
    para.remove_spans(keep_heading=True)


def test_insert_reference_after_element():
    para = Paragraph("abc")
    span = Span("s")
    para.append(span)
    para.insert_reference("ref1", after=span)
    assert span.xpath("text:reference-ref")


def test_insert_reference_text_format():
    para = Paragraph("abc")
    mark = ReferenceMarkStart("ref1")
    mark.set_attribute("text:name", "ref1")
    with patch.object(
        ReferenceMarkStart,
        "referenced_text",
        return_value="ref_text",
    ):
        doc = MagicMock()
        doc.get_reference_mark.return_value = mark
        with patch.object(
            Element, "document_body", new_callable=PropertyMock
        ) as mock_body:
            mock_body.return_value = doc
            para.insert_reference("ref1", ref_format="text")

    # Branch where get_reference_mark is not there
    with patch.object(
        Element,
        "document_body",
        new_callable=PropertyMock,
    ) as mock_body:
        mock_body.return_value = Element.from_tag("office:body")
        para.insert_reference("ref1", ref_format="text")


def test_set_bookmark_role_none():
    para = Paragraph("abc")
    para.set_bookmark("bm1", role=None, position=1)
    assert para.xpath("//text:bookmark")



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

from collections.abc import Iterable
from typing import cast
from unittest.mock import patch

import pytest

from odfdo.const import ODF_CONTENT, ODF_STYLES
from odfdo.content import Content
from odfdo.document import Document
from odfdo.element import Element
from odfdo.style import (
    Style,
    _make_line_string,
    _make_thick_string,
    make_table_cell_border_string,
)


@pytest.fixture
def content(samples) -> Iterable[Content]:
    document = Document(samples("span_style.odt"))
    content = cast(Content, document.get_part(ODF_CONTENT))
    yield content


def test_get_styles(content):
    styles = content.get_styles()
    assert len(styles) == 7


def test_get_styles_family(content):
    styles = content.get_styles(family="paragraph")
    assert len(styles) == 2


def test_get_style_automatic(content):
    style = content.get_style("paragraph", "P1")
    assert style is not None


def test_insert_style(content):
    style = Style(
        "paragraph",
        "style1",
        area="text",
        **{"fo:color": "#0000ff", "fo:background-color": "#ff0000"},
    )
    auto_styles = content.get_element("//office:automatic-styles")
    auto_styles.append(style)
    read_style = content.get_style("paragraph", "style1")

    expected = (
        "<style:style "
        'style:family="paragraph" style:name="style1">'
        "<style:text-properties "
        'fo:background-color="#ff0000" fo:color="#0000ff">'
        "</style:text-properties>"
        "</style:style>"
    )
    assert read_style._canonicalize() == expected


def test_get_styles_none(content):
    styles = content.get_styles(family=None)
    assert len(styles) == 7


def test_get_styles_nonexistent(content):
    with pytest.raises(ValueError, match="Unknown family"):
        content.get_styles(family="nonexistent")


def test_get_style_not_found(content):
    style = content.get_style("paragraph", "nonexistent")
    assert style is None


def test_get_style_by_element(content):
    style = content.get_style("paragraph", "P1")
    read_style = content.get_style("paragraph", style)
    assert read_style == style


def test_get_style_by_element_error(content):
    element = Element.from_tag("<text:p/>")
    with pytest.raises(ValueError, match="Not a odf_style"):
        content.get_style("paragraph", element)


def test_get_style_by_display_name(content):
    style = content.get_style("paragraph", "P1")
    style.set_attribute("style:display-name", "My Display Name")
    read_style = content.get_style(
        "paragraph",
        display_name="My Display Name",
    )
    assert read_style._canonicalize() == style._canonicalize()


def test_get_style_no_family(content):
    # get font face by name (it has style:name but no style:family)
    style = content.get_style("", "Liberation Serif")
    assert style.tag == "style:font-face"


def test_get_style_no_family_draw_name():
    # Test matching draw:name when family is empty
    container = Element.from_tag("<office:styles/>")
    # draw:fill-image uses draw:name
    tag = '<draw:fill-image draw:name="img1" xlink:href="url1" />'
    img = Element.from_tag(tag)
    container.append(img)

    style = container.get_style("", "img1")
    assert style.tag == "draw:fill-image"
    assert style.get_attribute("draw:name") == "img1"


def test_get_style_no_family_display_name():
    # Test matching display-name when family is empty
    container = Element.from_tag("<office:styles/>")
    tag = '<style:style style:name="s1" style:display-name="Display 1" />'
    s = Element.from_tag(tag)
    container.append(s)

    style = container.get_style("", display_name="Display 1")
    assert style.get_attribute("style:name") == "s1"


def test_get_style_no_family_first():
    # Test getting first style when family and names are empty
    container = Element.from_tag("<office:styles/>")
    tag1 = '<style:style style:name="s1" />'
    tag2 = '<style:style style:name="s2" />'
    container.append(Element.from_tag(tag1))
    container.append(Element.from_tag(tag2))

    style = container.get_style("")
    assert style.get_attribute("style:name") == "s1"


def test_get_style_default_by_family(samples):
    document = Document(samples("example.odt"))
    styles_part = document.get_part(ODF_STYLES)
    # Testing get_style(family, name=None) which should return default style
    style = styles_part.get_style("paragraph")
    assert style.tag == "style:default-style"
    assert style.get_attribute("style:family") == "paragraph"


def test_get_styles_all_no_family(samples):
    document = Document(samples("example.odt"))
    styles_part = document.get_part(ODF_STYLES)
    styles = styles_part.get_styles()  # family="" by default
    assert len(styles) > 0
    # Should include default styles, markers, fill-images etc.
    tags = {s.tag for s in styles}
    assert "style:default-style" in tags


def test_make_thick_string_variants():
    assert _make_thick_string(None) == "0.06pt"
    assert _make_thick_string("") == "0.06pt"
    assert _make_thick_string("  1pt  ") == "1pt"
    assert _make_thick_string(0.5) == "0.50pt"
    assert _make_thick_string(100) == "1.00pt"
    with pytest.raises(ValueError, match="Thickness must be None"):
        _make_thick_string([])


def test_make_line_string_variants():
    assert _make_line_string(None) == "solid"
    assert _make_line_string("") == "solid"
    assert _make_line_string("  dotted  ") == "dotted"
    with pytest.raises(ValueError, match="Line style must be None"):
        _make_line_string([])


def test_make_table_cell_border_string():
    res = make_table_cell_border_string(
        thick=1.0,
        line="dashed",
        color="blue",
    )
    assert res == "1.00pt dashed #0000FF"


def test_style_init_errors():
    obj = object.__new__(Style)
    with pytest.raises(TypeError, match="Wrong initializer"):
        obj.__init__(family="master-page")
    with pytest.raises(ValueError, match="Unknown family"):
        obj.__init__(family="unknown")


def test_style_init_font_face_error():
    obj = object.__new__(Style)
    with pytest.raises(ValueError, match="font_name is required"):
        obj.__init__(family="font-face")


def test_style_init_table_align_error():
    with pytest.raises(ValueError, match="Invalid align value"):
        Style(family="table", align="invalid")


def test_style_family_setter():
    style = Style(family="paragraph")
    style.family = "text"
    assert style.get_attribute("style:family") == "text"


def test_style_family_none_lazy():
    elem = Element.from_tag("text:list-style")
    style = Style(tag_or_elem=elem._Element__element)
    assert style.family == "list"


def test_style_get_level_style_non_list():
    style = Style(family="paragraph")
    assert style.get_level_style(1) is None


def test_style_set_level_style_non_list():
    style = Style(family="paragraph")
    assert style.set_level_style(1, num_format="1") is None


def test_style_set_level_style_unknown_type():
    style = Style(family="list")
    with pytest.raises(ValueError, match="unknown level style type"):
        style.set_level_style(1)


def test_style_set_level_style_existing_transmute():
    style = Style(family="list")
    style.set_level_style(1, bullet_char="*")
    style.set_level_style(1, num_format="1")
    level1 = style.get_level_style(1)
    assert level1.tag == "text:list-level-style-number"


def test_style_set_level_style_other_args():
    style = Style(family="list")
    style.set_level_style(
        1,
        num_format="1",
        display_levels=2,
        start_value=5,
        style="T1",
    )
    level1 = style.get_level_style(1)
    assert level1.get_attribute("text:display-levels") == "2"
    assert level1.get_attribute("text:start-value") == "5"
    assert level1.text_style == "T1"


def test_style_set_level_style_none_return():
    style = Style(family="list")
    with patch.object(Element, "from_tag", return_value=None):
        assert style.set_level_style(1, num_format="1") is None


def test_set_font_full():
    style = Style(
        family="font-face",
        font_name="Arial",
        font_family_generic="swiss",
    )
    assert style.font_family_generic == "swiss"

    style.set_font("Courier", family_generic="roman")
    assert style.font_family_generic == "roman"

    style.set_font("Verdana", family=None)
    assert style.svg_font_family == '"Verdana"'


def test_set_font_non_font_face():
    style = Style(family="paragraph")
    style.set_font("Arial")
    assert style.name is None


def test_style_repr():
    style = Style(family="paragraph", name="S1")
    assert "family=paragraph" in repr(style)
    assert "name=S1" in repr(style)


def test_style_init_various_families_missing_branches_1():
    s1 = Style("paragraph", area="text")
    props = s1.get_properties(area="text")
    assert props is None


def test_style_init_various_families_missing_branches_2():
    s2 = Style("paragraph", area="table-cell")
    props2 = s2.get_properties(area="table-cell")
    assert props2 is None


def test_style_init_various_families_missing_branches_3():
    s3 = Style("paragraph", area="table-row")
    assert s3.get_properties(area="table-row") is None


def test_style_init_various_families_missing_branches_4():
    s4 = Style("paragraph", area="table-column")
    assert s4.get_properties(area="table-column") is None


def test_style_init_various_families_missing_branches_5():
    s5 = Style("paragraph", area="table")
    assert s5.get_properties(area="table") is None


def test_style_init_various_families_missing_branches_6():
    s6 = Style("paragraph", area="graphic")
    assert s6.get_properties(area="graphic") is None


def test_style_init_various_families():
    s = Style(
        "paragraph",
        name="S1",
        display_name="D1",
        parent_style="P1",
        master_page="M1",
    )
    assert s.name == "S1"
    assert s.display_name == "D1"
    assert s.parent_style == "P1"
    assert s.master_page == "M1"


def test_style_init_various_families_1():
    s1 = Style(
        "text",
        italic=True,
        bold=True,
        color="blue",
        background_color="red",
    )
    props = s1.get_properties()
    assert props["fo:font-style"] == "italic"
    assert props["fo:font-weight"] == "bold"


def test_style_init_various_families_2():
    s2 = Style("table-cell", shadow="gray", background_color="yellow")
    props = s2.get_properties()
    assert props["style:shadow"] == "gray"
    assert props["fo:background-color"] == "#FFFF00"


def test_style_init_various_families_3():
    s3 = Style(
        "table-row",
        height="1cm",
        use_optimal_height=True,
        background_color="blue",
    )
    props = s3.get_properties()
    assert props["style:row-height"] == "1cm"
    assert props["style:use-optimal-row-height"] == "true"
    assert props["fo:background-color"] == "#0000FF"


def test_style_init_various_families_4():
    s4 = Style(
        "table-column",
        width="2cm",
        break_before="page",
        break_after="column",
    )
    props = s4.get_properties()
    assert props["style:column-width"] == "2cm"
    assert props["fo:break-before"] == "page"
    assert props["fo:break-after"] == "column"

    s_table = Style("table", width="15cm", align="center")
    props = s_table.get_properties()
    assert props["style:width"] == "15cm"
    assert props["table:align"] == "center"


def test_style_init_various_families_5():
    s5 = Style("graphic", min_height="3cm")
    props = s5.get_properties()
    assert props["fo:min-height"] == "3cm"


def test_style_set_level_style_url_and_clone():
    style = Style("list")
    style.set_level_style(1, url="img.png", prefix="(", suffix=")")
    level1 = style.get_level_style(1)
    assert level1.tag == "text:list-level-style-image"
    assert level1.get_attribute("xlink:href") == "img.png"
    assert level1.get_attribute("style:num-prefix") == "("
    assert level1.get_attribute("style:num-suffix") == ")"

    style2 = Style("list")
    style2.set_level_style(1, clone=level1)
    assert style2.get_level_style(1).get_attribute("xlink:href") == "img.png"


def test_style_init_font_face_full():
    s = Style(
        "font-face",
        font_name="MyFont",
        font_family="Serif",
        font_family_generic="roman",
        font_pitch="fixed",
    )
    assert s.name == "MyFont"
    assert s.svg_font_family == '"Serif"'
    assert s.font_family_generic == "roman"
    assert s.font_pitch == "fixed"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.shapes import DrawControl


@pytest.fixture
def control() -> Iterable[DrawControl]:
    shape = DrawControl(
        name="Name",
        control="control id",
        size=("10cm", "12cm"),
        position=("3cm", "4cm"),
    )
    yield shape


def test_draw_control_minimal():
    shape = DrawControl()
    assert shape._canonicalize() == "<draw:control></draw:control>"


def test_draw_control_class():
    shape = Element.from_tag("<draw:control/>")
    assert isinstance(shape, DrawControl)


def test_draw_control_control(control):
    assert control.control == "control id"


def test_draw_control_position_1(control):
    assert control.position == ("3cm", "4cm")


def test_draw_control_position_2(control):
    assert control.pos_x == "3cm"


def test_draw_control_position_3(control):
    assert control.pos_y == "4cm"


def test_draw_control_size_1(control):
    assert control.size == ("10cm", "12cm")


def test_draw_control_size_2(control):
    assert control.height == "12cm"


def test_draw_control_size_3(control):
    assert control.width == "10cm"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
from __future__ import annotations

from collections.abc import Iterable
from unittest.mock import patch

import pytest

from odfdo.cell import Cell
from odfdo.column import Column
from odfdo.document import Document
from odfdo.row import Row
from odfdo.table import Table
from odfdo.table_cache import TableCache


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield document.body.get_table(name="Example1")


def test_iter_columns(table):
    assert len(list(table.iter_columns())) == 7


def test_traverse_func(table):
    # compatibility
    assert table.traverse_columns == table.iter_columns


def test_traverse_columns(table):
    assert len(list(table.traverse_columns())) == 7


def test_get_column_list(table):
    assert len(list(table.get_columns())) == 7


def test_get_column_list_property(table):
    assert len(table.columns) == 7


def test_get_column_list_style(table):
    coordinates = [col.x for col in table.get_columns(style=r"co2")]
    assert coordinates == [2, 3]


def test_get_column(table):
    column = table.get_column(3)
    assert column.style == "co2"
    assert column.x == 3
    column = table.get_column(4)
    assert column.style == "co1"
    assert column.x == 4


def test_set_column(table):
    column = table.get_column(3)
    column_back = table.set_column(4, column)
    assert column_back.x == 4
    column = table.get_column(4)
    assert column.x == 4
    assert column.style == "co2"


def test_insert(table):
    column = table.insert_column(3)
    assert type(column) is Column
    assert column.x == 3


def test_insert_column(table):
    column = table.insert_column(3, Column())
    assert table.width == 8
    assert table.get_row(0).width == 8
    assert column.x == 3


def test_append(table):
    column = table.append_column()
    assert type(column) is Column
    assert column.x == table.width - 1


def test_append_column(table):
    column = table.append_column(Column())
    assert table.width == 8
    assert table.get_row(0).width == 7
    assert column.x == table.width - 1
    # The column must be inserted between the columns and the rows
    assert type(table.children[-1]) is not Column


def test_delete_column(table):
    table.delete_column(3)
    assert table.width == 6
    assert table.get_row(0).width == 6


def test_get_column_cell_values(table):
    assert table.get_column_values(3) == [2, 2, 2, 4]


def test_set_column_cell_values(table):
    table.set_column_values(5, ["a", "b", "c", "d"])
    assert table.get_values() == [
        [1, 1, 1, 2, 3, "a", 3],
        [1, 1, 1, 2, 3, "b", 3],
        [1, 1, 1, 2, 3, "c", 3],
        [1, 2, 3, 4, 5, "d", 7],
    ]


def test_is_column_empty():
    table = Table("Empty", width=10, height=20)
    for x in range(10):
        assert table.is_column_empty(x) is True


def test_is_column_empty_no():
    table = Table("Not Empty", width=10, height=20)
    table.set_value((4, 9), "Bouh !")
    assert table.is_column_empty(4) is False


def test_get_column_values_filter():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("A2", 123)
    # vals: complete=True, get_type=True
    vals = table.get_column_values(
        0,
        cell_type="string",
        complete=True,
        get_type=True,
    )
    assert len(vals) == 2
    assert vals[0] == ("v1", "string")
    assert vals[1] == (None, None)

    # vals: complete=True, get_type=False
    vals = table.get_column_values(
        0,
        cell_type="string",
        complete=True,
        get_type=False,
    )
    assert vals[1] is None

    # vals: complete=False, get_type=True
    vals = table.get_column_values(
        0,
        cell_type="string",
        complete=False,
        get_type=True,
    )
    assert len(vals) == 1


def test_get_column_values_more():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("A2", 123)
    cell_float = table.get_cell("A2", clone=False)
    with patch.object(Table, "get_column_cells", return_value=[cell_float]):
        vals = table.get_column_values(
            0,
            cell_type="string",
            complete=True,
            get_type=False,
        )
        assert vals == [None]
        vals = table.get_column_values(
            0,
            cell_type="string",
            complete=True,
            get_type=True,
        )
        assert vals == [(None, None)]


def test_get_column_values_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "get_column_cells", return_value=[None]):
        vals = table.get_column_values(0, complete=True, get_type=False)
        assert vals == [None]
    with patch.object(Table, "get_column_cells", return_value=[None]):
        vals = table.get_column_values(0, complete=False)
        assert vals == []


def test_get_column_values_branch_gaps():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("A2", 123)
    # Case 1: cell is NOT None but cell_type mismatch
    c2 = table.get_cell("A2")
    with patch.object(Table, "get_column_cells", return_value=[c2]):
        assert table.get_column_values(
            0, cell_type="string", complete=True, get_type=True
        ) == [(None, None)]
        assert table.get_column_values(
            0, cell_type="string", complete=True, get_type=False
        ) == [None]


def test_get_column_values_branch_complete_get_type_false():
    table = Table("T")
    table.set_value("A1", "v1")
    # Case mismatch AND complete=True AND get_type=False
    with patch.object(Table, "get_column_cells", return_value=[table.get_cell("A1")]):
        assert table.get_column_values(
            0, cell_type="float", complete=True, get_type=False
        ) == [None]


def test_get_column_values_complete_none_branch():
    table = Table("T")
    table.set_value("A1", "v1")
    with patch.object(Table, "get_column_cells", return_value=[None]):
        res = table.get_column_values(0, complete=True, get_type=False)
        assert res == [None]
        res = table.get_column_values(0, complete=True, get_type=True)
        assert res == [(None, None)]


def test_get_column_values_branch_complete_get_type_true_mismatch():
    table = Table("T")
    table.set_value("A1", "v1")
    # Case mismatch AND complete=True AND get_type=True
    with patch.object(Table, "get_column_cells", return_value=[table.get_cell("A1")]):
        assert table.get_column_values(
            0, cell_type="float", complete=True, get_type=True
        ) == [(None, None)]


def test_get_column_values_complete_branches_hits():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("A2", "v2")
    # c1 is string "v1". Let's filter by "float"
    c1 = table.get_cell("A1")
    with patch.object(Table, "get_column_cells", return_value=[c1]):
        res = table.get_column_values(
            0, cell_type="float", complete=True, get_type=True
        )
        assert res == [(None, None)]

        res = table.get_column_values(
            0, cell_type="float", complete=True, get_type=False
        )
        assert res == [None]

        res = table.get_column_values(0, cell_type="float", complete=False)
        assert res == []


def test_iter_columns_params():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    table.set_value("C1", "v3")
    assert list(table.iter_columns(start=2, end=1)) == []
    cols = list(table.iter_columns(start=1, end=1))
    assert len(cols) == 1
    cols = list(table.iter_columns(start=-1))
    assert len(cols) == 3


def test_get_columns_coord():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    table.set_value("C1", "v3")
    cols = table.get_columns(coord="B:C")
    assert len(cols) == 2


def test_get_column2_edge():
    table = Table("Test")
    col = table._get_column2(10)
    assert col.tag == "table:table-column"


def test_delete_column_outside():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.delete_column(10)
    assert table.width == 1


def test_get_column_cells_filter():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("A2", "v2")
    cell1 = table.get_cell("A1", clone=False)
    cell1.style = "s1"
    cells = table.get_column_cells(0, style="s1")
    assert len(cells) == 1
    cells = table.get_column_cells(0, content="v2")
    assert len(cells) == 1
    cells = table.get_column_cells(0, cell_type="string")
    assert len(cells) == 2
    cells = table.get_column_cells(0, style="s1", complete=True)
    assert len(cells) == 2
    assert cells[0].style == "s1"
    assert cells[1] is None

    # test mismatched content
    cells = table.get_column_cells(0, content="wrong", complete=True)
    assert len(cells) == 2
    assert cells[0] is None
    assert cells[1] is None


def test_get_column_cells_none_error():
    table = Table("Test")
    table.set_value("A1", "v1")
    # Using mock to force get_cell to return None
    with patch.object(Row, "get_cell", return_value=None):
        with pytest.raises(ValueError):
            table.get_column_cells(0, cell_type="string")


def test_set_column_none():
    table = Table("Test")
    col = table.set_column(0, column=None)
    assert col.tag == "table:table-column"


def test_set_column_append():
    table = Table("Test")
    _col = table.set_column(0, Column())
    assert table.width == 1


def test_set_column_far_append():
    table = Table("Test")
    _col = table.set_column(2, Column())
    assert table.width == 3


def test_insert_column_edges():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.insert_column(1, Column())
    assert table.width == 2
    table.insert_column(5, Column())
    assert table.width == 6
    table.insert_column(0, Column())
    assert table.get_value("B1") == "v1"
    assert table.width == 7


def test_is_column_empty_missing_cell():
    table = Table("Test")
    table.set_value("A1", "v1")
    assert table.is_column_empty(0) is False
    assert table.is_column_empty(1) is True


def test_append_column_repeated():
    table = Table("Test")
    col = Column(repeated=2)
    table.append_column(col, _repeated=2)
    assert table.width == 2


def test_set_column_cells_wrong_height():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("A2", "v2")
    # height is 2
    with pytest.raises(ValueError, match="col mismatch"):
        table.set_column_cells(0, [Cell()])


def test_delete_column_uneven_branch_hits_final():
    table = Table("T")
    # width is 0.
    table.set_value("A1", "v")
    table.set_value("B1", "v")
    # table.width is now 2.

    # Add a row 1 with 1 cell only
    table.set_value("A2", "v")
    # row 1 width is 1. table.width is 2.

    # delete column 1 (B)
    # width = 2
    table.delete_column(1)
    assert table.get_value("A1") == "v"
    assert table.get_value("A2") == "v"


def test_delete_column_branch_part_false_hits():
    # Table width 3, row 1 width 1.
    table = Table("T")
    table.set_value("A1", "v")
    table.set_value("B1", "v")
    table.set_value("C1", "v")
    # table width 3.
    table.set_value("A2", "v")
    # row 1 width 1.
    # delete column 2 (C).
    # width becomes 2.
    table.delete_column(2)
    assert table.get_value("A2") == "v"


def test_get_column_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_element_idx2", return_value=None):
        table._table_cache.col_elements = {}
        # Restore width
        table._table_cache.col_map = [0]
        with pytest.raises(ValueError):
            table.get_column(0)


def test_delete_column_uneven_rows():
    table = Table("Uneven")
    table.set_value("A2", "v2")  # row 1 width 1
    table.set_value("B1", "v1")  # row 0 width 2
    # delete column 1 (B). row 1 width is 1. x=1. 1 > 1 is False.
    table.delete_column(1)
    assert table.get_value("A1") is None
    assert table.get_value("A2") == "v2"


def test_is_column_empty_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    # Force get_column_cells to return [None]
    with patch.object(Table, "get_column_cells", return_value=[None]):
        assert table.is_column_empty(0) is True


def test_get_column2_none_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    # We patch the CLASS TableCache
    with patch.object(TableCache, "col_idx", return_value=None):
        assert table._get_column2(0) is None
    with patch.object(Table, "_get_element_idx2", return_value=None):
        table._table_cache.col_elements = {}
        # Restore width
        table._table_cache.col_map = [0]
        assert table._get_column2(0) is None


def test_insert_column_uneven_rows():
    table = Table("Uneven")
    table.set_value("A1", "v1")  # row 0 width 1
    table.set_value("B2", "v2")  # row 1 width 2
    table.insert_column(1)
    assert table.width == 3
    assert table.get_value("A1") == "v1"
    assert table.get_value("C2") == "v2"


def test_append_column_none_internal_2():
    table = Table("Test")
    table.set_value("A1", "v1")
    with patch.object(Table, "_get_element_idx2", return_value=None):
        table._table_cache.col_elements = {}
        # Restore width
        table._table_cache.col_map = [0]
        with pytest.raises(ValueError):
            table.append_column(Column())


def test_delete_column_get_value():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("B2", "v2")
    # row 0 width is 1. row 1 width is 2.
    # delete column 1 (B). row 0 width is 1. x is 1. 1 > 1 is False.
    table.delete_column(1)
    # row 0 was skipped by if row.width > x
    assert table.get_value("A1") == "v1"
    # row 1 had B2 deleted, so C2 became B2? No, B2 is at x=1.
    # Actually B2 was at x=1. row 1 width was 2. 2 > 1 is True.
    # row 1.delete_cell(1) was called.
    assert table.get_value("A2") is None


def test_get_cells_none_branch_part():
    table = Table("T")
    table.set_value("A1", "v")
    # Coverage for get_cells branch where some cell is None
    # We need a row where get_cell returns None
    with patch.object(Row, "get_cell", return_value=None):
        with pytest.raises(ValueError):
            table.get_column_cells(0, cell_type="string")


def test_get_column2_cache_miss_xml_exists():
    table = Table("T")
    table.set_value("A1", "v")
    table._table_cache.col_elements = {}
    # ensure it's in the col_map
    assert 0 in table._table_cache.col_map
    col = table._get_column2(0)
    assert col.tag == "table:table-column"


def test_delete_column_shorter_row_branch():
    table = Table("T")
    table.set_value("A1", "v")
    table.set_value("B1", "v")
    # row 0 width 2
    table.set_value("A2", "v")
    # row 1 width 1
    # total width 2
    # delete column 1 (B). row 1 width is 1. width is 2. 1 >= 2 is False.
    table.delete_column(1)
    assert table.get_value("A1") == "v"
    assert table.get_value("A2") == "v"


def test_get_column2_cached_branch():
    table = Table("T")
    table.set_value("A1", "v")
    # Cache it
    _col = table._get_column2(0)
    col2 = table._get_column2(0)
    assert col2.tag == "table:table-column"


def test_get_column_cells_none_mock_branch():
    table = Table("T")
    table.set_value("A1", "v")
    with patch.object(Row, "get_cell", return_value=None):
        with pytest.raises(ValueError):
            table.get_column_cells(0, cell_type="string")


def test_insert_column_shorter_row_branch():
    table = Table("T")
    table.set_value("A1", "v")
    # width 1
    table.set_value("A2", "v")
    # insert at 5. row 0 width 1. 1 > 5 is False.
    table.insert_column(5)
    assert table.width == 6


def test_delete_column_uneven_branch_hits():
    table = Table("T")
    table.set_value("A1", "v")
    table.set_value("B1", "v")
    # row 0 width is 2.
    table.set_value("A2", "v")
    # row 1 width is 1.
    # Table width is 2.
    # delete column 1 (B).
    table.delete_column(1)
    assert table.get_value("A1") == "v"
    assert table.get_value("B1") is None
    assert table.get_value("A2") == "v"



# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from unittest.mock import PropertyMock, patch

import pytest

from odfdo.cell import Cell
from odfdo.document import Document
from odfdo.element import Element
from odfdo.frame import Frame
from odfdo.image import DrawImage
from odfdo.table import Table


def test_set_cell_image_malformed():
    table = Table("T")
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        with pytest.raises(ValueError):
            table.set_cell_image("1", frame, doc_type="spreadsheet")
    with pytest.warns(DeprecationWarning):
        with pytest.raises(ValueError):
            table.set_cell_image("A", frame, doc_type="spreadsheet")


def test_set_cell_image_paragraph_error():
    table = Table("T")
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)

    with patch.object(Cell, "get_element", return_value=None):
        with pytest.warns(DeprecationWarning):
            with pytest.raises(ValueError):
                table.set_cell_image("A1", frame, doc_type="text")


def test_set_cell_image_no_body_internal():
    table = Table("T")
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    with patch.object(Table, "document_body", new_callable=PropertyMock) as mock_body:
        mock_body.return_value = None
        with pytest.warns(DeprecationWarning):
            with pytest.raises(ValueError, match="document type not found"):
                table.set_cell_image("A1", frame)


def test_set_cell_image_unsupported_body_internal():
    table = Table("T")
    doc = Document("ods")
    table = doc.body.get_table(0)
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    # We mock body.tag to return something not in dict
    with patch.object(Table, "document_body", new_callable=PropertyMock) as mock_body:
        body = doc.body
        with patch.object(Element, "tag", new_callable=PropertyMock) as mock_tag:
            mock_tag.return_value = "office:other"
            mock_body.return_value = body
            with pytest.warns(DeprecationWarning):
                with pytest.raises(
                    ValueError, match="document type not supported for images"
                ):
                    table.set_cell_image("A1", frame)


def test_set_cell_image_doc_type_text_branch():
    table = Table("T")
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="text")
    cell = table.get_cell("A1")
    assert cell.get_element("text:p/draw:frame") is not None


def test_set_cell_image_spreadsheet_branch():
    table = Table("T")
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="spreadsheet")
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is not None


def test_set_cell_image_text_branch_hit():
    table = Table("T")
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="text")
    cell = table.get_cell("A1")
    assert cell.get_element("text:p/draw:frame") is not None


def test_set_cell_image_spreadsheet_branch_hit():
    table = Table("T")
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="spreadsheet")
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is not None


def test_set_cell_image_auto_text_branch_hit():
    doc = Document("odt")
    table = Table("T")
    doc.body.append(table)
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    # Trigger 1791 branch part (True) via auto doc_type="text"
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame)
    cell = table.get_cell("A1")
    assert cell.get_element("text:p/draw:frame") is not None


def test_set_cell_image_auto_spreadsheet_branch_hit():
    doc = Document("ods")
    table = doc.body.get_table(0)
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    # Trigger 1791 branch part (False) via auto doc_type="spreadsheet"
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame)
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is not None


def test_set_cell_image_skip_both_branches():
    table = Table("T")
    img = DrawImage("test.png")

    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="other")
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is None


def test_set_cell_image_spreadsheet():
    doc = Document("ods")
    table = doc.body.get_table(0)
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="spreadsheet")
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is not None


def test_set_cell_image_text():
    doc = Document("odt")
    table = Table("Test")
    doc.body.append(table)
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    table.set_value("A1", "")
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame, doc_type="text")
    cell = table.get_cell("A1")
    assert cell.get_element("text:p/draw:frame") is not None


def test_set_cell_image_no_doctype():
    doc = Document("ods")
    table = doc.body.get_table(0)
    img = DrawImage("test.png")
    frame = Frame.image_frame(img)
    with pytest.warns(DeprecationWarning):
        table.set_cell_image("A1", frame)
    cell = table.get_cell("A1")
    assert cell.get_element("draw:frame") is not None



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import os
import platform
import runpy
import subprocess
import sys
from pathlib import Path

import pytest

from odfdo import Document
from odfdo.scripts import flat
from odfdo.scripts.flat import main as main_script
from odfdo.scripts.flat import main_convert_flat, parse_cli_args
from odfdo.scripts.folder import main_convert_folder
from odfdo.scripts.folder import parse_cli_args as folder_parse_cli_args

SCRIPT = Path(flat.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_flat_no_param():
    params = []
    out, err, exitcode = run_params(params)
    print(out, err, exitcode)
    assert exitcode == 2
    if platform.system() != "Windows":
        assert "odfdo-flat: error: the following arguments are required" in err
        assert "usage" in err


# direct access to internal function


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Fails on Windows due to stdin/output capture conflict.",
)
def test_flat_2_no_param_on_main_function(monkeypatch):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.value.code >= 1


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="Fails on Windows due to stdin/output capture conflict.",
)
def test_flat_2_no_param():
    with pytest.raises(SystemExit) as result:
        params = parse_cli_args([])
        main_convert_flat(params)
        assert result.value.code >= 1


def test_flat_2_version(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--version"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "odfdo-flat v3" in captured.out


def test_flat_2_help(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--help"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "Convert a standard ODF file" in captured.out


def test_flat_2_no_file_on_main(monkeypatch):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", ["xxx", "none_file"])
        main_script()
        assert result.value.code >= 1


def test_flat_2_no_file():
    params = parse_cli_args(["none_file"])
    with pytest.raises(SystemExit) as result:
        main_convert_flat(params)
        assert result.value.code >= 1


def test_flat_2_odt_to_fodt(capsys, tmp_path, samples):
    source = samples("test_diff1.odt")
    document_src = Document(source)
    tmp_source = tmp_path / "test.odt"
    document_src.save(tmp_source)
    assert tmp_source.exists()
    params = parse_cli_args([str(tmp_source)])
    main_convert_flat(params)
    captured = capsys.readouterr()
    print(captured.out)
    print(captured.err)
    path = tmp_path / "test.fodt"
    assert path.is_file()


def test_flat_2_fodt_to_odt(tmp_path, samples):
    source = samples("test_diff1.odt")
    document_src = Document(source)
    tmp_source = tmp_path / "test.odt"
    document_src.save(tmp_source)
    # First convert to flat
    params1 = parse_cli_args([str(tmp_source)])
    main_convert_flat(params1)
    tmp_source.unlink()
    assert not tmp_source.exists()
    # Then convert back to ODF
    params2 = parse_cli_args([str(tmp_path / "test.fodt")])
    main_convert_flat(params2)
    assert tmp_source.exists()


def test_flat_2_ods_to_fods(capsys, tmp_path, samples):
    source = samples("simple_table.ods")
    document_src = Document(source)
    tmp_source = tmp_path / "test.ods"
    document_src.save(tmp_source)
    assert tmp_source.exists()
    params = parse_cli_args([str(tmp_source)])
    main_convert_flat(params)
    captured = capsys.readouterr()
    print(captured.out)
    print(captured.err)
    path = tmp_path / "test.fods"
    assert path.is_file()


def test_flat_2_fods_to_ods(tmp_path, samples):
    source = samples("simple_table.ods")
    document_src = Document(source)
    tmp_source = tmp_path / "test.ods"
    document_src.save(tmp_source)
    # First convert to flat
    params1 = parse_cli_args([str(tmp_source)])
    main_convert_flat(params1)
    tmp_source.unlink()
    assert not tmp_source.exists()
    # Then convert back to ODF
    params2 = parse_cli_args([str(tmp_path / "test.fods")])
    main_convert_flat(params2)
    assert tmp_source.exists()


def test_flat_2_folder_to_fodt(capsys, tmp_path, samples):
    # First create a folder structure
    source = samples("test_diff1.odt")
    document_src = Document(source)
    tmp_source = tmp_path / "test.odt"
    document_src.save(tmp_source)
    # Convert to folder

    params1 = folder_parse_cli_args([str(tmp_source)])
    main_convert_folder(params1)
    tmp_source.unlink()
    assert not tmp_source.exists()
    # Convert folder to flat
    params2 = parse_cli_args([str(tmp_path / "test.odt.folder")])
    main_convert_flat(params2)
    captured = capsys.readouterr()
    print(captured.out)
    print(captured.err)
    path = tmp_path / "test.fodt"
    assert path.is_file()


def test_is_flat_xml_file_1(tmp_path):
    # Test with recognized suffix
    p1 = tmp_path / "test.fodt"
    p1.write_text("dummy")
    assert flat.is_flat_xml_file(p1) is True


def test_is_flat_xml_file_2(tmp_path):
    # Test with .xml suffix and correct content
    p2 = tmp_path / "test.xml"
    p2.write_text(
        '<?xml version="1.0" encoding="UTF-8"?>'
        "<office:document xmlns:office="
        '"urn:oasis:names:tc:opendocument:xmlns:office:1.0"/>'
    )
    assert flat.is_flat_xml_file(p2) is True


def test_is_flat_xml_file_3(tmp_path):
    # Test with no suffix and correct content
    p3 = tmp_path / "test_no_suffix"
    p3.write_text(
        '<?xml version="1.0" encoding="UTF-8"?>'
        "<office:document xmlns:office="
        '"urn:oasis:names:tc:opendocument:xmlns:office:1.0"/>'
    )
    assert flat.is_flat_xml_file(p3) is True


def test_is_flat_xml_file_4(tmp_path):
    # Test with .xml suffix but wrong content
    p4 = tmp_path / "wrong.xml"
    p4.write_text('<?xml version="1.0"?><root/>')
    assert flat.is_flat_xml_file(p4) is False


def test_is_flat_xml_file_5(tmp_path):
    # Test with no suffix and wrong content
    p5 = tmp_path / "wrong_no_suffix"
    p5.write_text("not xml at all")
    assert flat.is_flat_xml_file(p5) is False


def test_is_flat_xml_file_6(tmp_path):
    # Test with other suffix
    p6 = tmp_path / "test.txt"
    p6.write_text(
        '<?xml version="1.0" encoding="UTF-8"?>'
        "<office:document xmlns:office="
        '"urn:oasis:names:tc:opendocument:xmlns:office:1.0"/>'
    )
    assert flat.is_flat_xml_file(p6) is False


def test_is_flat_xml_file_oserror(tmp_path):
    # Trigger OSError by passing a directory where a file is expected
    p = tmp_path / "test_dir.xml"
    p.mkdir()
    assert flat.is_flat_xml_file(p) is False


def test_convert_flat_not_file_or_folder(tmp_path):
    fifo = tmp_path / "test_fifo"
    try:
        os.mkfifo(fifo)
    except (AttributeError, OSError):
        pytest.skip("FIFOs not supported on this platform")

    with pytest.raises(ValueError, match="Not a file or folder"):
        flat.convert_flat(str(fifo))


def test_flat_runpy(tmp_path, samples, monkeypatch):
    source = samples("test_diff1.odt")
    tmp_source = tmp_path / "test_runpy.odt"
    tmp_source.write_bytes(source.read_bytes())
    monkeypatch.setattr(sys, "argv", ["odfdo-flat", str(tmp_source)])
    runpy.run_path(str(SCRIPT), run_name="__main__")
    assert (tmp_path / "test_runpy.fodt").exists()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_properties import FormListProperty


@pytest.fixture
def form_list_property() -> Iterable[FormListProperty]:
    yield FormListProperty(
        property_name="some name",
        value_type="string",
    )


def test_form_list_property_class():
    form = FormListProperty()
    assert isinstance(form, FormListProperty)


def test_form_list_property_minimal_tag():
    form = Element.from_tag("<form:list-property/>")
    assert isinstance(form, FormListProperty)


def test_form_list_property_serialize():
    form = FormListProperty()
    assert form.serialize() == "<form:list-property/>"


def test_form_list_property_property_name(form_list_property):
    assert form_list_property.property_name == "some name"


def test_form_list_property_value_type(form_list_property):
    assert form_list_property.value_type == "string"
